# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

import raillabel
from raillabel.format import annotation_classes

# == Fixtures =========================

@pytest.fixture
def all_annotations(
    bbox, bbox_train,
    cuboid,
    poly2d,
    poly3d,
    seg3d,
):
    return {
        bbox.uid: bbox,
        bbox_train.uid: bbox_train,
        cuboid.uid: cuboid,
        poly2d.uid: poly2d,
        poly3d.uid: poly3d,
        seg3d.uid: seg3d,
    }

# == Tests ============================

def test_post_init_happy(object_person, point2d, size2d):
    raillabel.format.Bbox(
        uid="d2764400-8560-4991-a491-ada598b345c8",
        object=object_person,
        pos=point2d,
        size=size2d,
    )

def test_post_init_unhappy(object_person, point2d):
    with pytest.raises(TypeError):
        raillabel.format.Bbox(
            uid="d2764400-8560-4991-a491-ada598b345c8",
            object=object_person,
            pos=point2d,
        )


def test_annotation_classes():
    assert annotation_classes() == {
        "bbox": raillabel.format.Bbox,
        "poly2d": raillabel.format.Poly2d,
        "cuboid": raillabel.format.Cuboid,
        "poly3d": raillabel.format.Poly3d,
        "vec": raillabel.format.Seg3d,
    }


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import json
import os
import sys
import tempfile
from copy import deepcopy
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent))

import raillabel


def test_save_scene(json_paths):
    with tempfile.TemporaryDirectory("w") as temp_dir:

        scene_orig = raillabel.load(json_paths["openlabel_v1_short"], False, False)

        raillabel.save(scene_orig, Path(temp_dir) / "test_save_file.json")
        scene_saved = raillabel.load(Path(temp_dir) / "test_save_file.json", False, False)

    assert scene_orig == scene_saved


def test_save_json(json_data):
    with tempfile.TemporaryDirectory("w") as temp_dir:

        stripped_input_data = deepcopy(json_data["openlabel_v1_short"])

        # Removes the object data pointers from the example file so that it needs to be generated from the data
        for object in stripped_input_data["openlabel"]["objects"].values():
            del object["frame_intervals"]
            del object["object_data_pointers"]

        with (Path(temp_dir) / "stripped_input_data.json").open("w") as f:
            json.dump(stripped_input_data, f)

        scene = raillabel.load(Path(temp_dir) / "stripped_input_data.json", False, False)
        raillabel.save(scene, Path(temp_dir) / "test_save_file.json")

        with (Path(temp_dir) / "test_save_file.json").open() as f:
            saved_and_loaded_data = json.load(f)

    # Removes the exporter version from the generated file as these are hard to test for
    if "exporter_version" in saved_and_loaded_data["openlabel"]["metadata"]:
        del saved_and_loaded_data["openlabel"]["metadata"]["exporter_version"]

    assert saved_and_loaded_data == json_data["openlabel_v1_short"]


def test_frame_intervals():
    data = {
        "openlabel": {
            "metadata": {"schema_version": "1.0.0"},
            "frames": {
                "0": {},
                "1": {},
                "2": {},
                "5": {},
                "7": {},
                "8": {},
            },
        }
    }

    scene = raillabel.load_.loader_classes.LoaderRailLabel().load(data)
    dict_repr = scene.asdict()["openlabel"]

    assert "frame_intervals" in dict_repr
    assert dict_repr["frame_intervals"] == [
        {"frame_start": 0, "frame_end": 2},
        {"frame_start": 5, "frame_end": 5},
        {"frame_start": 7, "frame_end": 8},
    ]


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from .test_attributes import (
    attributes_multiple_types,
    attributes_multiple_types_dict,
    attributes_single_type,
    attributes_single_type_dict,
)
from .test_bbox import bbox, bbox_dict, bbox_train, bbox_train_dict
from .test_cuboid import cuboid, cuboid_dict
from .test_element_data_pointer import (
    element_data_pointer_full,
    element_data_pointer_full_dict,
    element_data_pointer_minimal,
    element_data_pointer_minimal_dict,
)
from .test_frame import frame, frame_dict
from .test_frame_interval import frame_interval, frame_interval_dict
from .test_intrinsics_pinhole import intrinsics_pinhole, intrinsics_pinhole_dict
from .test_intrinsics_radar import intrinsics_radar, intrinsics_radar_dict
from .test_metadata import (
    metadata_full,
    metadata_full_dict,
    metadata_minimal,
    metadata_minimal_dict,
)
from .test_num import num, num_dict
from .test_object import (
    object_person,
    object_person_dict,
    object_train,
    object_train_dict,
    objects,
    objects_dict,
)
from .test_object_annotation import all_annotations
from .test_object_data import object_data_person_dict, object_data_train_dict
from .test_point2d import point2d, point2d_another, point2d_another_dict, point2d_dict
from .test_point3d import point3d, point3d_another, point3d_another_dict, point3d_dict
from .test_poly2d import poly2d, poly2d_dict
from .test_poly3d import poly3d, poly3d_dict
from .test_quaternion import quaternion, quaternion_dict
from .test_scene import scene, scene_dict
from .test_seg3d import seg3d, seg3d_dict
from .test_sensor import (
    coordinate_systems_dict,
    sensor_camera,
    sensor_camera_dict,
    sensor_lidar,
    sensor_lidar_dict,
    sensor_radar,
    sensor_radar_dict,
    sensors,
    streams_dict,
)
from .test_sensor_reference import sensor_reference_camera, sensor_reference_camera_dict
from .test_size2d import size2d, size2d_dict
from .test_size3d import size3d, size3d_dict
from .test_transform import transform, transform_dict



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass

from _collections_abc import dict_values

from .element_data_pointer import AttributeType, ElementDataPointer
from .frame_interval import FrameInterval

if t.TYPE_CHECKING:
    from .frame import Frame


@dataclass
class Object:
    """Physical, unique object in the data, that can be tracked via its UID.

    Parameters
    ----------
    uid: str
        This a string representing the unique universal identifier for the object.
    name: str
        Name of the object. It is a friendly name and not used for indexing. Commonly the class
        name is used followed by an underscore and an integer (i.e. person_0032).
    type: str
        The type of an object defines the class the object corresponds to.
    """

    uid: str
    name: str
    type: str

    # --- Public Methods --------------

    @classmethod
    def fromdict(cls, data_dict: dict, object_uid: str) -> "Object":
        """Generate a Object from a dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        object_uid: str
            Unique identifier of the object.

        Returns
        -------
        object: raillabel.format.Object
            Converted object.
        """
        return Object(uid=object_uid, type=data_dict["type"], name=data_dict["name"])

    def asdict(self, frames: t.Optional[t.Dict[int, "Frame"]] = None) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.
        frames: dict, optional
            The dictionary of frames stored under Scene.frames used for the frame intervals and
            object data pointers. If None, these are not provided. Default is None.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        if frames is None:
            return {"name": str(self.name), "type": str(self.type)}

        else:
            return {
                "name": str(self.name),
                "type": str(self.type),
                "frame_intervals": self._frame_intervals_asdict(self.frame_intervals(frames)),
                "object_data_pointers": self._object_data_pointers_asdict(
                    self.object_data_pointers(frames)
                ),
            }

    def frame_intervals(self, frames: t.Dict[int, "Frame"]) -> t.List[FrameInterval]:
        """Return frame intervals in which this object is present.

        Parameters
        ----------
        frames: dict[int, raillabel.format.Frame]
            The dictionary of frames stored under Scene.frames.

        Returns
        -------
        list[FrameInterval]
            List of the FrameIntervals, where this object is contained.
        """

        frame_uids_containing_object = [
            frame.uid for frame in frames.values() if self._is_object_in_frame(frame)
        ]

        return FrameInterval.from_frame_uids(frame_uids_containing_object)

    def object_data_pointers(self, frames: t.Dict[int, "Frame"]) -> t.Dict[str, ElementDataPointer]:
        """Create object data pointers used in WebLABEL visualization.

        Parameters
        ----------
        frames: dict[int, raillabel.format.Frame]
            The dictionary of frames stored under Scene.frames.

        Returns
        -------
        dict[str, ElementDataPointer]
            ObjectDataPointers dict as required by WebLABEL. Keys are the ObjectDataPointer uids.
        """

        pointer_ids_per_frame = self._collect_pointer_ids_per_frame(frames)
        frame_uids_per_pointer_id = self._reverse_frame_pointer_ids(pointer_ids_per_frame)
        frame_intervals_per_pointer_id = self._convert_to_intervals(frame_uids_per_pointer_id)

        attributes_per_pointer_id = self._collect_attributes_per_pointer_id(frames)
        attribute_pointers_per_pointer_id = self._convert_to_attribute_pointers(
            attributes_per_pointer_id
        )

        return self._create_object_data_pointers(
            frame_intervals_per_pointer_id, attribute_pointers_per_pointer_id
        )

    # --- Private Methods -------------

    def _frame_intervals_asdict(self, frame_intervals: t.List[FrameInterval]) -> dict:
        return [fi.asdict() for fi in frame_intervals]

    def _object_data_pointers_asdict(
        self, object_data_pointers: t.Dict[str, ElementDataPointer]
    ) -> dict:
        return {
            pointer_id: pointer.asdict() for pointer_id, pointer in object_data_pointers.items()
        }

    def _is_object_in_frame(self, frame: "Frame") -> bool:
        return self.uid in frame.object_data

    def _filtered_annotations(self, frame: "Frame") -> dict_values:
        return [ann for ann in frame.annotations.values() if ann.object.uid == self.uid]

    def _collect_pointer_ids_per_frame(
        self, frames: t.Dict[int, "Frame"]
    ) -> t.Dict[int, t.Set[str]]:

        pointer_ids_per_frame = {}
        for frame in frames.values():
            pointer_ids_per_frame[frame.uid] = set()

            for annotation in self._filtered_annotations(frame):
                pointer_ids_per_frame[frame.uid].add(annotation.name)

        return pointer_ids_per_frame

    def _reverse_frame_pointer_ids(
        self, pointer_ids_per_frame: t.Dict[int, t.Set[str]]
    ) -> t.Dict[str, t.Set[int]]:

        frame_uids_per_pointer_id = {}
        for frame_uid, pointer_ids in pointer_ids_per_frame.items():
            for pointer_id in pointer_ids:

                if pointer_id not in frame_uids_per_pointer_id:
                    frame_uids_per_pointer_id[pointer_id] = set()

                frame_uids_per_pointer_id[pointer_id].add(frame_uid)

        return frame_uids_per_pointer_id

    def _convert_to_intervals(
        self, frame_uids_per_pointer_id: t.Dict[str, t.Set[int]]
    ) -> t.Dict[str, t.List[FrameInterval]]:

        frame_intervals = {}
        for pointer_id, frame_uids in frame_uids_per_pointer_id.items():
            frame_intervals[pointer_id] = FrameInterval.from_frame_uids(list(frame_uids))

        return frame_intervals

    def _collect_attributes_per_pointer_id(
        self, frames: t.Dict[int, "Frame"]
    ) -> t.Dict[str, t.Dict[str, t.Any]]:

        attributes_per_pointer_id = {}
        for frame in frames.values():
            for annotation in self._filtered_annotations(frame):
                if annotation.name not in attributes_per_pointer_id:
                    attributes_per_pointer_id[annotation.name] = {}

                attributes_per_pointer_id[annotation.name].update(annotation.attributes)

        return attributes_per_pointer_id

    def _convert_to_attribute_pointers(
        self, attributes_per_pointer_id: t.Dict[str, t.Dict[str, t.Any]]
    ) -> t.Dict[str, t.Dict[str, AttributeType]]:
        for attributes in attributes_per_pointer_id.values():
            for attribute_name, attribute_value in attributes.items():
                attributes[attribute_name] = AttributeType.from_value(type(attribute_value))

        return attributes_per_pointer_id

    def _create_object_data_pointers(
        self,
        frame_intervals_per_pointer_id: t.Dict[str, t.List[FrameInterval]],
        attribute_pointers_per_pointer_id: t.Dict[str, t.Dict[str, AttributeType]],
    ) -> t.Dict[str, ElementDataPointer]:

        object_data_pointers = {}
        for pointer_id in frame_intervals_per_pointer_id:
            object_data_pointers[pointer_id] = ElementDataPointer(
                uid=pointer_id,
                frame_intervals=frame_intervals_per_pointer_id[pointer_id],
                attribute_pointers=attribute_pointers_per_pointer_id[pointer_id],
            )

        return object_data_pointers

    # --- Special Methods -------------

    def __hash__(self) -> int:
        """Return hash."""
        return self.uid.__hash__()



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass


@dataclass
class Quaternion:
    """A quaternion.

    Parameters
    ----------
    x: float or int
        The x component of the quaternion.
    y: float or int
        The y component of the quaternion.
    z: float or int
        The z component of the quaternion.
    w: float or int
        The w component of the quaternion.
    """

    x: float
    y: float
    z: float
    w: float

    @classmethod
    def fromdict(cls, data_dict: dict) -> "Quaternion":
        """Generate a Quaternion object from a dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        """

        return Quaternion(
            x=data_dict[0],
            y=data_dict[1],
            z=data_dict[2],
            w=data_dict[3],
        )

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        return [float(self.x), float(self.y), float(self.z), float(self.w)]



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import json
import typing as t
from pathlib import Path

from ... import exceptions, format
from ..._util._warning import _WarningsLogger
from ._loader_abc import LoaderABC


class LoaderRailLabel(LoaderABC):
    """Loader class for the OpenLabel v1 annotation format.

    Attributes
    ----------
    scene: raillabel.Scene
        Loaded raillabel.Scene with the data.
    warnings: t.List[str]
        List of warning strings, that have been found during the execution of load().
    """

    scene: format.Scene
    warnings: t.List[str]

    SCHEMA_PATH: Path = (
        Path(__file__).parent.parent.parent / "validate" / "schemas" / "raillabel_schema.json"
    )

    @property
    def subschema_version(self):
        """Return subschema version."""
        with self.SCHEMA_PATH.open() as schema_file:
            subschema_version = json.load(schema_file)["version"]
        return subschema_version

    def load(self, data: dict, validate: bool = False) -> format.Scene:
        """Load the data into a raillabel.Scene and return it.

        Parameters
        ----------
        data: dict
            A dictionary loaded from a JSON-file.
        validate: bool, optional
            If True, the annotation data is validated via the respective schema. This is
            recommended, if you are working with a modified or non-official file. Setting this
            option will increase loading time. Default is False.

        Returns
        -------
        scene: raillabel.Scene
            The loaded scene with the data.

        Raises
        ------
        raillabel.exceptions.SchemaError
            if validate is True and the data does not validate against the schema.
        """

        if validate:
            self.validate(data)

        with _WarningsLogger() as logger:
            self.scene = format.Scene.fromdict(data, self.subschema_version)

        self.warnings = logger.warnings

        return self.scene

    def supports(self, data: dict) -> bool:
        """Determine if the loader is suitable for the data (lightweight).

        Parameters
        ----------
        data: dict
            A dictionary loaded from a JSON-file.

        Returns
        -------
        bool:
            If True, the Loader class is suitable for the data.
        """

        if "openlabel" not in data or "metadata" not in data["openlabel"]:
            return False

        if "subschema_version" in data["openlabel"]["metadata"]:
            return (
                "openlabel" in data
                and "metadata" in data["openlabel"]
                and "schema_version" in data["openlabel"]["metadata"]
                and data["openlabel"]["metadata"]["subschema_version"].split(".")[0] in ["2", "3"]
            )

        else:
            return (
                "openlabel" in data
                and "metadata" in data["openlabel"]
                and "schema_version" in data["openlabel"]["metadata"]
            )

    def _prepare_data(self, data: dict) -> dict:
        """Add optional fields to dict to simplify interaction.

        Parameters
        ----------
        data : dict
            JSON data.

        Returns
        -------
        dict
            Enhanced JSON data.
        """

        if "coordinate_systems" not in data["openlabel"]:
            data["openlabel"]["coordinate_systems"] = {}

        if "streams" not in data["openlabel"]:
            data["openlabel"]["streams"] = {}

        if "objects" not in data["openlabel"]:
            data["openlabel"]["objects"] = {}

        if "frames" not in data["openlabel"]:
            data["openlabel"]["frames"] = {}

        return data["openlabel"]

    def _check_sensor_completeness(self, cs_data: dict, stream_data: dict):
        """Check for corresponding cs and stream completeness.

        Parameters
        ----------
        cs_data : dict
            Coordinate system data in the RailLabel format.
        stream_data : dict
            Stream data in the RailLabel format.

        Raises
        ------
        raillabel.exceptions.MissingCoordinateSystemError
            if a stream has no corresponding coordinate system.
        raillabel.exceptions.MissingStreamError
            if a coordinate system has no corresponding stream.
        raillabel.exceptions.UnsupportedParentError
            if a coordinate system has no corresponding stream.
        """

        for stream_uid in stream_data:
            if stream_uid not in cs_data:
                raise exceptions.MissingCoordinateSystemError(
                    f"Stream {stream_uid} has no corresponding coordinate system."
                )

        for cs_uid in cs_data:
            if cs_uid == "base":
                continue

            if cs_data[cs_uid]["parent"] != "base":
                raise exceptions.UnsupportedParentError(
                    f"Only 'base' is permitted as a parent for coordinate system {cs_uid}, "
                    + f"not {cs_data[cs_uid]['parent']}."
                )

            if cs_uid not in stream_data:
                raise exceptions.MissingStreamError(
                    f"Coordinate sytem {cs_uid} has no corresponding stream."
                )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass


@dataclass
class Size3d:
    """Dimensional information of an object in 3d.

    Parameters
    ----------
    width: float
        Size of the object in the x-dimension.
    length: float
        Size of the object in the y-dimension.
    height: float
        Size of the object in the z-dimension.
    """

    width: float
    length: float
    height: float

    @classmethod
    def fromdict(cls, data_dict: dict) -> "Size3d":
        """Generate a Size3d from a dictionary in the UAI format.

        Parameters
        ----------
        data_dict: dict
            Understand.AI T4 format dictionary containing the data_dict.

        Returns
        -------
        Size3d
            Converted 3d size.
        """

        return Size3d(
            width=float(data_dict["width"]),
            length=float(data_dict["length"]),
            height=float(data_dict["height"]),
        )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

from raillabel._util._clean_dict import _clean_dict


def test_clean_dict():
    input_dict = {
        "non_empty_field": "non_empty_value",
        "none_field": None,
        "field_with_len_0": []
    }

    assert _clean_dict(input_dict) == {
        "non_empty_field": "non_empty_value",
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel


def test_valid_file(json_data):
    validation_result = raillabel.validate(json_data["openlabel_v1_short"])
    assert validation_result[0]


def test_file_one_type_error(json_data):
    data = json_data["openlabel_v1_short"]

    data["openlabel"]["streams"]["lidar"]["uri"] = 42

    validation_result = raillabel.validate(data)

    assert not validation_result[0] and len(validation_result[1]) == 1


def test_file_two_type_errors(json_data):
    data = json_data["openlabel_v1_short"]

    data["openlabel"]["streams"]["lidar"]["uri"] = 42
    data["openlabel"]["coordinate_systems"]["base"]["type"] = "invalid_value"

    validation_result = raillabel.validate(data)

    assert not validation_result[0] and len(validation_result[1]) == 2


def test_valid_file_path(json_data, json_paths):
    validation_result = raillabel.validate(json_data["openlabel_v1_short"], str(json_paths["raillabel_schema"]))
    assert validation_result[0]


def test_invalid_file_path(json_data, json_paths):
    json_paths["raillabel_schema"] = str(json_paths["raillabel_schema"]) + "_invalid"

    with pytest.raises(FileNotFoundError):
        raillabel.validate(json_data["openlabel_v1_short"], json_paths["raillabel_schema"])


def test_invalid_schema_key(json_data):
    with pytest.raises(FileNotFoundError):
        raillabel.validate(json_data["openlabel_v1_short"], "invalid_schema")


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import Metadata

# == Fixtures =========================

@pytest.fixture
def metadata_minimal_dict() -> dict:
    return {
        "schema_version": "1.0.0"
    }


@pytest.fixture
def metadata_full_dict() -> dict:
    return {
        "schema_version": "1.0.0",
        "annotator": "test_annotator",
        "subschema_version": "2.1.0",
        "comment": "test_comment",
        "name": "test_project",
        "tagged_file": "test_folder",
    }

@pytest.fixture
def metadata_minimal() -> dict:
    return Metadata(
        schema_version="1.0.0"
    )


@pytest.fixture
def metadata_full() -> dict:
    return Metadata(
        schema_version="1.0.0",
        annotator="test_annotator",
        subschema_version="2.1.0",
        comment="test_comment",
        name="test_project",
        tagged_file="test_folder",
    )

# == Tests ============================

def test_fromdict_minimal():
    metadata = Metadata.fromdict(
        {
            "schema_version": "1.0.0"
        },
    )

    assert metadata.schema_version == "1.0.0"
    assert metadata.annotator is None


def test_fromdict_full():
    metadata = Metadata.fromdict(
        {
            "schema_version": "1.0.0",
            "annotator": "test_annotator",
            "subschema_version": "2.1.0",
            "comment": "test_comment",
            "name": "test_project",
            "tagged_file": "test_folder",
        },
        "2.1.1"
    )

    assert metadata.annotator == "test_annotator"
    assert metadata.schema_version == "1.0.0"
    assert metadata.comment == "test_comment"
    assert metadata.name == "test_project"
    assert metadata.subschema_version == "2.1.1"
    assert metadata.tagged_file == "test_folder"


def test_fromdict_additional_arg_valid():
    metadata = Metadata.fromdict(
        {
            "schema_version": "1.0.0",
            "additional_argument": "Some Value"
        }
    )

    assert metadata.schema_version == "1.0.0"
    assert metadata.additional_argument == "Some Value"


def test_fromdict_additional_arg_invalid():
    with pytest.raises(KeyError):
        Metadata.fromdict(
            {
                "schema_version": "1.0.0",
                "invalid python variable": "Some Value"
            }
        )


def test_asdict_minimal():
    metadata_dict = Metadata(
        schema_version="1.0.0"
    ).asdict()

    assert metadata_dict == {
        "schema_version": "1.0.0"
    }


def test_asdict_full():
    metadata_dict = Metadata(
        annotator="test_annotator",
        schema_version="1.0.0",
        comment="test_comment",
        name="test_project",
        subschema_version="2.1.0",
        tagged_file="test_folder",
    ).asdict()

    assert metadata_dict == {
        "schema_version": "1.0.0",
        "annotator": "test_annotator",
        "subschema_version": "2.1.0",
        "comment": "test_comment",
        "name": "test_project",
        "tagged_file": "test_folder",
    }


def test_fromdict_additional_arg():
    metadata = Metadata(
        schema_version="1.0.0"
    )

    metadata.additional_argument = "Some Value"

    assert metadata.asdict() == {
        "schema_version": "1.0.0",
        "additional_argument": "Some Value"
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from pathlib import Path

from ..._util._warning import _WarningsLogger
from ...format import understand_ai as uai_format
from ._loader_abc import LoaderABC
from .loader_raillabel import LoaderRailLabel


class LoaderUnderstandAi(LoaderABC):
    """Loader class for the Understand.Ai Trains4 annotation format.

    Attributes
    ----------
    scene: raillabel.format.understand_ai.Scene
        Loaded raillabel.format.understand_ai.Scene with the data.
    warnings: t.List[str]
        List of warning strings, that have been found during the execution of load().
    """

    scene: uai_format.Scene
    warnings: t.List[str]

    SCHEMA_PATH: Path = (
        Path(__file__).parent.parent.parent
        / "validate"
        / "schemas"
        / "understand_ai_t4_schema.json"
    )

    def load(self, data: dict, validate: bool = False) -> uai_format.Scene:
        """Load the data into a UAIScene and return it.

        Parameters
        ----------
        data: dict
            A dictionary loaded from a JSON-file.
        validate: bool
            If True, the annotation data is validated via the respective schema. This is highly
            recommended, as not validating the data may lead to Errors during loading or while
            handling the scene. However, validating may increase the loading time. Default is False.

        Returns
        -------
        scene: raillabel.format.understand_ai.UAIScene
            The loaded scene with the data.
        """

        if validate:
            self.validate(data)

        with _WarningsLogger() as logger:
            data_converted_to_raillabel = uai_format.Scene.fromdict(data).to_raillabel()

        raillabel_loader = LoaderRailLabel()
        raillabel_scene = raillabel_loader.load(data_converted_to_raillabel, validate=False)

        self.warnings = logger.warnings + raillabel_loader.warnings

        return raillabel_scene

    def supports(self, data: dict) -> bool:
        """Determine if the loader is suitable for the data (lightweight).

        Parameters
        ----------
        data: dict
            A dictionary loaded from a JSON-file.

        Returns
        -------
        bool:
            If True, the Loader class is suitable for the data.
        """

        return (
            "metadata" in data
            and "project_id" in data["metadata"]
            and "coordinateSystems" in data
            and "frames" in data
        )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass
from uuid import UUID

from ._annotation import _Annotation
from .sensor_reference import SensorReference


@dataclass
class Segmentation3d(_Annotation):
    """The 3D segmentation of a lidar pointcloud.

    Parameters
    ----------
    id: uuid.UUID
        Unique identifier of the annotation.
    object_id: uuid.UUID
        Unique identifier of the object this annotation refers to. Used for tracking.
    class_name: str
        Name of the class this annotation belongs to.
    attributes: dict[str, str or list]
        Key value pairs of attributes with the keys beeing the friendly identifier of the
        attribute and the value beeing the attribute value.
    sensor: raillabel.format.understand_ai.SensorReference
        Information about the sensor this annotation is labeled in.
    associated_points: list[int]
        List of point indices of the lidar pointcloud.
    number_of_points: int
        Total number of points in the associated_points.
    """

    associated_points: t.List[int]
    number_of_points: int

    OPENLABEL_ID = "vec"

    @classmethod
    def fromdict(cls, data_dict: t.Dict) -> "Segmentation3d":
        """Generate a Segmentation3d from a dictionary in the UAI format.

        Parameters
        ----------
        data_dict: dict
            Understand.AI T4 format dictionary containing the data_dict.

        Returns
        -------
        Segmentation3d
            Converted 3d segmentation.
        """

        return Segmentation3d(
            id=UUID(data_dict["id"]),
            object_id=UUID(data_dict["objectId"]),
            class_name=data_dict["className"],
            attributes=data_dict["attributes"],
            sensor=SensorReference.fromdict(data_dict["sensor"]),
            associated_points=data_dict["geometry"]["associatedPoints"],
            number_of_points=data_dict["geometry"]["numberOfPointsInBox"],
        )

    def _val_to_raillabel(self) -> list:
        return self.associated_points



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel import exceptions
from raillabel._util._warning import _WarningsLogger
from raillabel.format import Frame, FrameInterval, Scene

# == Fixtures =========================

@pytest.fixture
def scene_dict(
    metadata_full_dict,
    sensor_camera_dict, sensor_lidar_dict, sensor_radar_dict,
    object_person_dict, object_train_dict,
    frame, frame_dict,
) -> dict:
    return {
        "openlabel": {
            "metadata": metadata_full_dict,
            "streams": {
                sensor_camera_dict["uid"]: sensor_camera_dict["stream"],
                sensor_lidar_dict["uid"]: sensor_lidar_dict["stream"],
                sensor_radar_dict["uid"]: sensor_radar_dict["stream"],
            },
            "coordinate_systems": {
                "base": {
                    "type": "local",
                    "parent": "",
                    "children": [
                        sensor_lidar_dict["uid"],
                        sensor_camera_dict["uid"],
                        sensor_radar_dict["uid"],
                    ]
                },
                sensor_camera_dict["uid"]: sensor_camera_dict["coordinate_system"],
                sensor_lidar_dict["uid"]: sensor_lidar_dict["coordinate_system"],
                sensor_radar_dict["uid"]: sensor_radar_dict["coordinate_system"],
            },
            "objects": {
                object_person_dict["object_uid"]: object_person_dict["data_dict"],
                object_train_dict["object_uid"]: object_train_dict["data_dict"],
            },
            "frames": {
                frame.uid: frame_dict,
            },
            "frame_intervals": [
                {
                    "frame_start": 0,
                    "frame_end": 0,
                }
            ]
        }
    }

@pytest.fixture
def scene(
    metadata_full,
    sensor_camera, sensor_lidar, sensor_radar,
    object_person, object_train,
    frame
) -> Scene:
    return Scene(
        metadata=metadata_full,
        sensors={
            sensor_lidar.uid: sensor_lidar,
            sensor_camera.uid: sensor_camera,
            sensor_radar.uid: sensor_radar,
        },
        objects={
            object_person.uid: object_person,
            object_train.uid: object_train,
        },
        frames={frame.uid: frame}
    )

# == Tests ============================

def test_fromdict_metadata(
    metadata_full, metadata_full_dict,
):
    scene = Scene.fromdict(
        {
            "openlabel": {
                "metadata": metadata_full_dict,
            }
        },
        subschema_version=metadata_full.subschema_version,
    )

    scene.metadata.exporter_version = None  # necessary for testing on remote

    assert scene.metadata == metadata_full

def test_fromdict_sensors(
    metadata_full_dict,
    sensor_camera, sensor_lidar, sensor_radar,
    sensor_camera_dict, sensor_lidar_dict, sensor_radar_dict,
):
    scene = Scene.fromdict(
        {
            "openlabel": {
                "metadata": metadata_full_dict,
                "streams": {
                    sensor_camera_dict["uid"]: sensor_camera_dict["stream"],
                    sensor_lidar_dict["uid"]: sensor_lidar_dict["stream"],
                    sensor_radar_dict["uid"]: sensor_radar_dict["stream"],
                },
                "coordinate_systems": {
                    "base": {
                        "type": "local",
                        "parent": "",
                        "children": [
                            sensor_lidar_dict["uid"],
                            sensor_camera_dict["uid"],
                            sensor_radar_dict["uid"],
                        ]
                    },
                    sensor_camera_dict["uid"]: sensor_camera_dict["coordinate_system"],
                    sensor_lidar_dict["uid"]: sensor_lidar_dict["coordinate_system"],
                    sensor_radar_dict["uid"]: sensor_radar_dict["coordinate_system"],
                },
            }
        }
    )

    assert scene.sensors == {
        sensor_lidar.uid: sensor_lidar,
        sensor_camera.uid: sensor_camera,
        sensor_radar.uid: sensor_radar,
    }

def test_fromdict_missing_coordinate_system(
    metadata_full_dict,
    sensor_camera_dict, sensor_lidar_dict, sensor_radar_dict,
):
    with pytest.raises(exceptions.MissingCoordinateSystemError):
        Scene.fromdict(
            {
                "openlabel": {
                    "metadata": metadata_full_dict,
                    "streams": {
                        sensor_camera_dict["uid"]: sensor_camera_dict["stream"],
                        sensor_lidar_dict["uid"]: sensor_lidar_dict["stream"],
                    },
                    "coordinate_systems": {
                        "base": {
                            "type": "local",
                            "parent": "",
                            "children": [
                                sensor_camera_dict["uid"],
                            ]
                        },
                        sensor_camera_dict["uid"]: sensor_camera_dict["coordinate_system"],
                    },
                }
            }
        )

def test_fromdict_missing_stream(
    metadata_full_dict,
    sensor_camera_dict, sensor_lidar_dict, sensor_radar_dict,
):
    with pytest.raises(exceptions.MissingStreamError):
        Scene.fromdict(
            {
                "openlabel": {
                    "metadata": metadata_full_dict,
                    "streams": {
                        sensor_camera_dict["uid"]: sensor_camera_dict["stream"],
                    },
                    "coordinate_systems": {
                        "base": {
                            "type": "local",
                            "parent": "",
                            "children": [
                                sensor_lidar_dict["uid"],
                                sensor_camera_dict["uid"],
                            ]
                        },
                        sensor_camera_dict["uid"]: sensor_camera_dict["coordinate_system"],
                        sensor_lidar_dict["uid"]: sensor_lidar_dict["coordinate_system"],
                    },
                }
            }
        )

def test_fromdict_unsupported_parent(
    metadata_full_dict,
    sensor_camera_dict, sensor_lidar_dict, sensor_radar_dict,
):
    sensor_camera_dict["coordinate_system"]["parent"] = "unsupported_parent"

    with pytest.raises(exceptions.UnsupportedParentError):
        Scene.fromdict(
            {
                "openlabel": {
                    "metadata": metadata_full_dict,
                    "streams": {
                        sensor_camera_dict["uid"]: sensor_camera_dict["stream"],
                    },
                    "coordinate_systems": {
                        "base": {
                            "type": "local",
                            "parent": "",
                            "children": [
                                sensor_camera_dict["uid"],
                            ]
                        },
                        sensor_camera_dict["uid"]: sensor_camera_dict["coordinate_system"],
                    },
                }
            }
        )

def test_fromdict_objects(
    metadata_full, metadata_full_dict,
    object_person, object_train,
    object_person_dict, object_train_dict,
):
    scene = Scene.fromdict(
        {
            "openlabel": {
                "metadata": metadata_full_dict,
                "objects": {
                    object_person_dict["object_uid"]: object_person_dict["data_dict"],
                    object_train_dict["object_uid"]: object_train_dict["data_dict"],
                },
            }
        },
        subschema_version=metadata_full.subschema_version,
    )

    assert scene.objects == {
        object_person.uid: object_person,
        object_train.uid: object_train,
    }

def test_fromdict_frames(
    metadata_full, metadata_full_dict,
    streams_dict, coordinate_systems_dict,
    objects_dict,
    frame, frame_dict,
):
    scene = Scene.fromdict(
        {
            "openlabel": {
                "metadata": metadata_full_dict,
                "streams": streams_dict,
                "coordinate_systems": coordinate_systems_dict,
                "objects": objects_dict,
                "frames": {
                    str(frame.uid): frame_dict,
                },
                "frame_intervals": [
                    {
                        "frame_start": 0,
                        "frame_end": 0,
                    }
                ]
            }
        },
        subschema_version=metadata_full.subschema_version,
    )

    assert scene.frames == {
         frame.uid: frame,
    }

def test_fromdict_duplicate_frame_id(metadata_full_dict):
    with _WarningsLogger() as logger:
        scene = Scene.fromdict(
            {
                "openlabel": {
                    "metadata": metadata_full_dict,
                    "frames": {
                        "0": {},
                        "00": {},
                    }
                }
            }
        )

    assert len(logger.warnings) == 1
    assert "0" in logger.warnings[0]
    assert len(scene.frames) == 1


def test_asdict_sensors(
    metadata_full, metadata_full_dict,
    sensor_camera, sensor_lidar, sensor_radar,
    sensor_camera_dict, sensor_lidar_dict, sensor_radar_dict,
):
    scene = Scene(
        metadata=metadata_full,
        sensors={
            sensor_lidar.uid: sensor_lidar,
            sensor_camera.uid: sensor_camera,
            sensor_radar.uid: sensor_radar,
        }
    )

    assert scene.asdict() == {
        "openlabel": {
            "metadata": metadata_full_dict,
            "streams": {
                sensor_camera_dict["uid"]: sensor_camera_dict["stream"],
                sensor_lidar_dict["uid"]: sensor_lidar_dict["stream"],
                sensor_radar_dict["uid"]: sensor_radar_dict["stream"],
            },
            "coordinate_systems": {
                "base": {
                    "type": "local",
                    "parent": "",
                    "children": [
                        sensor_lidar_dict["uid"],
                        sensor_camera_dict["uid"],
                        sensor_radar_dict["uid"],
                    ]
                },
                sensor_camera_dict["uid"]: sensor_camera_dict["coordinate_system"],
                sensor_lidar_dict["uid"]: sensor_lidar_dict["coordinate_system"],
                sensor_radar_dict["uid"]: sensor_radar_dict["coordinate_system"],
            },
        }
    }

def test_asdict_objects(
    metadata_full, metadata_full_dict,
    object_person, object_train,
    object_person_dict, object_train_dict,
):
    scene = Scene(
        metadata=metadata_full,
        objects={
            object_person.uid: object_person,
            object_train.uid: object_train,
        },
    )

    assert scene.asdict(calculate_pointers=False) == {
        "openlabel": {
            "metadata": metadata_full_dict,
            "objects": {
                object_person_dict["object_uid"]: object_person_dict["data_dict"],
                object_train_dict["object_uid"]: object_train_dict["data_dict"],
            },
        }
    }

def test_asdict_frames(
    metadata_full, metadata_full_dict,
    sensors, streams_dict, coordinate_systems_dict,
    objects, objects_dict,
    frame, frame_dict,
):
    scene = Scene(
        metadata=metadata_full,
        sensors=sensors,
        objects=objects,
        frames={
            frame.uid: frame,
        }
    )

    assert scene.asdict(calculate_pointers=False) == {
        "openlabel": {
            "metadata": metadata_full_dict,
            "streams": streams_dict,
            "coordinate_systems": coordinate_systems_dict,
            "objects": objects_dict,
            "frames": {
                str(frame.uid): frame_dict,
            },
            "frame_intervals": [
                {
                    "frame_start": 0,
                    "frame_end": 0,
                }
            ]
        }
    }


def test_frame_intervals(metadata_minimal):
    scene = Scene(
        metadata=metadata_minimal,
        frames={
            1: Frame(1),
            2: Frame(2),
            3: Frame(3),
            8: Frame(8),
        }
    )

    assert scene.frame_intervals == [
        FrameInterval(1, 3),
        FrameInterval(8, 8),
    ]


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Devkit for working with recorded and annotated train ride data from DB."""
from importlib import metadata

from . import _util, format, stats
from .exceptions import *
from .filter.filter import filter
from .format import Scene
from .load_.load import load
from .save.save import save
from .validate.validate import validate

try:
    __version__ = metadata.version("raillabel")
except metadata.PackageNotFoundError:
    __version__ = "0.0.0+unknown"
del metadata



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from enum import Enum


class AttributeType(Enum):
    """Enum of all valid RailLabel attribute types."""

    TEXT = "text"
    NUM = "num"
    BOOLEAN = "boolean"
    VEC = "vec"

    @classmethod
    def from_value(cls, attribute_value_class: t.Type) -> "AttributeType":
        """Return AttributeType based on class of attribute value.

        Parameters
        ----------
        attribute_value_class: type
            Class of the attribute value. Can be gathered by calling type()-function.

        Returns
        -------
        AttributeType
            Corresponding AttributeType.

        Raises
        ------
        ValueError
            if attribute value class does not correspond to an Attribute Type.
        """

        if attribute_value_class == str:
            return AttributeType.TEXT

        elif attribute_value_class in [float, int]:
            return AttributeType.NUM

        elif attribute_value_class == bool:
            return AttributeType.BOOLEAN

        elif attribute_value_class in [list, tuple]:
            return AttributeType.VEC

        else:
            raise ValueError(
                f"Type {attribute_value_class} does not correspond to a valid RailLabel attribute "
                + "type. Supported types are str, float, int, bool, list, tuple."
            )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import glob
import json
import sys
import typing as t
from pathlib import Path

import json5
import pytest

sys.path.insert(1, str(Path(__file__).parent.parent))
import raillabel

json_data_directories = [
    Path(__file__).parent / "__test_assets__",
    Path(__file__).parent.parent / "raillabel" / "validate" / "schemas"
]

@pytest.fixture(scope="session", autouse=True)
def compile_uncommented_test_file():
    """Compiles the main test file from json5 to json."""

    json5_file_path = json_data_directories[0] / "openlabel_v1_short.json5"

    with json5_file_path.open() as f:
        data = json5.load(f)

    with open(str(json5_file_path)[:-1], "w") as f:
        json.dump(data, f, indent=4)

@pytest.fixture
def json_paths(request) -> t.Dict[str, Path]:
    json_paths = _fetch_json_paths_from_cache(request)

    if json_paths is None:
        json_paths = {_get_file_identifier(p): p for p in _collect_json_paths()}

    return json_paths

def _fetch_json_paths_from_cache(request) -> t.Optional[t.Dict[str, Path]]:
    return request.config.cache.get("json_paths", None)

def _collect_json_paths() -> t.List[Path]:
    json_paths = []

    for dir in json_data_directories:
        json_paths.extend([Path(p) for p in glob.glob(str(dir) + "/**/**.json", recursive=True)])

    return json_paths

def _get_file_identifier(path: Path) -> str:
    """Return relative path from test asset dir as string."""

    if "__test_assets__" not in path.parts:
        return path.stem

    test_assets_dir_index = path.parts.index("__test_assets__")

    relative_path = ""
    for part in path.parts[test_assets_dir_index+1:-1]:
        relative_path += part + "/"

    relative_path += path.stem

    return relative_path

@pytest.fixture
def json_data(request) -> t.Dict[str, dict]:
    json_data = _fetch_json_data_from_cache(request)

    if json_data is None:
        json_data = {_get_file_identifier(p): _load_json_data(p) for p in _collect_json_paths()}

    return json_data

def _fetch_json_data_from_cache(request) -> t.Optional[t.Dict[str, Path]]:
    return request.config.cache.get("json_data", None)

def _load_json_data(path: Path) -> dict:
    with path.open() as f:
        json_data = json.load(f)
    return json_data



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass

from ._object_annotation import _ObjectAnnotation
from .object import Object
from .point2d import Point2d
from .size2d import Size2d


@dataclass
class Bbox(_ObjectAnnotation):
    """A 2D bounding box in an image.

    Parameters
    ----------
    uid: str
        This a string representing the unique universal identifier of the annotation.
    pos: raillabel.format.Point2d
        The center point of the bbox in pixels.
    size: raillabel.format.Size2d
        The dimensions of the bbox in pixels from the top left corner to the bottom right corner.
    object: raillabel.format.Object
        A reference to the object, this annotation belongs to.
    sensor: raillabel.format.Sensor, optional
        A reference to the sensor, this annotation is labeled in. Default is None.
    attributes: dict, optional
        Attributes of the annotation. Dict keys are the name str of the attribute, values are the
        attribute values. Default is {}.

    Properties (read-only)
    ----------------------
    name: str
        Name of the annotation used by the VCD player for indexing in the object data pointers.
    """

    pos: Point2d = None
    size: Size2d = None

    OPENLABEL_ID = "bbox"
    _REQ_FIELDS = ["pos", "size"]

    @classmethod
    def fromdict(cls, data_dict: dict, sensors: dict, object: Object) -> "Bbox":
        """Generate a Bbox object from a dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        sensors: dict
            Dictionary containing all sensors for the scene.
        object: raillabel.format.Object
            Object this annotation belongs to.

        Returns
        -------
        annotation: Bbox
            Converted annotation.
        """

        return Bbox(
            uid=str(data_dict["uid"]),
            pos=Point2d(x=data_dict["val"][0], y=data_dict["val"][1]),
            size=Size2d(x=data_dict["val"][2], y=data_dict["val"][3]),
            object=object,
            sensor=cls._coordinate_system_fromdict(data_dict, sensors),
            attributes=cls._attributes_fromdict(data_dict),
        )

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = self._annotation_required_fields_asdict()

        dict_repr["val"] = [
            float(self.pos.x),
            float(self.pos.y),
            float(self.size.x),
            float(self.size.y),
        ]

        dict_repr.update(self._annotation_optional_fields_asdict())

        return dict_repr



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass

from ._translation import fetch_sensor_resolutions, fetch_sensor_type, translate_sensor_id


@dataclass
class CoordinateSystem:
    """Global information for a sensor regarding calibration.

    Parameters
    ----------
    uid: str
        Friendly name of the sensor as well as its identifier. Must be unique
    topic: str
        Rostopic of the sensor.
    frame_id: str
        Name of the directory containing the files from this sensor.
    position: list of float
        3D translation with regards to the origin.
    rotation_quaternion: list of float
        Rotation quaternion with regards to the origin.
    rotation_matrix: list of float
        Rotation matrix with regards to the origin.
    angle_axis_rotation: list of float
        Angle axis rotation with regards to the origin.
    homogeneous_transform: list of float, optional
        Homogeneous transformation matrix with regards to the origin. Default is None.
    measured_position: list of float, optional
    camera_matrix: list of float, optional
        Camera matrix of the sensor. Only applies to sensors of type camera. Default is None.
    dist_coeffs: list of float, optional
        Distortion coefficients of the sensor. Only applies to sensors of type camera. Default is
        None.
    """

    uid: str
    topic: str
    frame_id: str
    position: t.List[float]
    rotation_quaternion: t.List[float]
    rotation_matrix: t.List[float]
    angle_axis_rotation: t.List[float]
    homogeneous_transform: t.Optional[t.List[float]] = None
    measured_position: t.Optional[t.List[float]] = None
    camera_matrix: t.Optional[t.List[float]] = None
    dist_coeffs: t.Optional[t.List[float]] = None

    @property
    def translated_uid(self) -> str:
        """Return uid translated to raillabel."""
        return translate_sensor_id(self.uid)

    @classmethod
    def fromdict(cls, data_dict: dict) -> "CoordinateSystem":
        """Generate a CoordinateSystem from a dictionary in the UAI format.

        Parameters
        ----------
        data_dict: dict
            Understand.AI T4 format dictionary containing the data.

        Returns
        -------
        coordinate_system: CoordinateSystem
            Converted coordinate_system.
        """

        return CoordinateSystem(
            uid=data_dict["coordinate_system_id"],
            topic=data_dict["topic"],
            frame_id=data_dict["frame_id"],
            position=data_dict["position"],
            rotation_quaternion=data_dict["rotation_quaternion"],
            rotation_matrix=data_dict["rotation_matrix"],
            angle_axis_rotation=data_dict["angle_axis_rotation"],
            homogeneous_transform=data_dict.get("homogeneous_transform"),
            measured_position=data_dict.get("measured_position"),
            camera_matrix=data_dict.get("camera_matrix"),
            dist_coeffs=data_dict.get("dist_coeffs"),
        )

    def to_raillabel(self) -> t.Tuple[dict, dict]:
        """Convert to a raillabel compatible dict.

        Returns
        -------
        coordinate_system_dict: dict
            Dictionary of the raillabel coordinate system.
        stream_dict: dict
            Dictionary of the raillabel stream.
        """

        stream_dict = {
            "type": "sensor",
            "parent": "base",
            "pose_wrt_parent": {
                "translation": self.position,
                "quaternion": self.rotation_quaternion,
            },
        }

        coordinate_system_dict = {
            "type": fetch_sensor_type(self.translated_uid),
            "uri": self.topic,
            "stream_properties": self._stream_properties_to_raillabel(
                fetch_sensor_type(self.translated_uid)
            ),
        }

        if coordinate_system_dict["stream_properties"] is None:
            del coordinate_system_dict["stream_properties"]

        return stream_dict, coordinate_system_dict

    def _stream_properties_to_raillabel(self, type: str) -> t.Optional[dict]:

        if type == "camera":
            return {
                "intrinsics_pinhole": {
                    "camera_matrix": self._convert_camera_matrix(self.camera_matrix[:]),
                    "distortion_coeffs": self.dist_coeffs,
                    "width_px": fetch_sensor_resolutions(self.translated_uid)["x"],
                    "height_px": fetch_sensor_resolutions(self.translated_uid)["y"],
                }
            }

        elif type == "radar":
            return {
                "intrinsics_radar": {
                    "resolution_px_per_m": fetch_sensor_resolutions(self.translated_uid)[
                        "resolution_px_per_m"
                    ],
                    "width_px": fetch_sensor_resolutions(self.translated_uid)["x"],
                    "height_px": fetch_sensor_resolutions(self.translated_uid)["y"],
                }
            }

        else:
            return None

    def _convert_camera_matrix(self, camera_matrix: list) -> list:

        camera_matrix.insert(9, 0)
        camera_matrix.insert(6, 0)
        camera_matrix.insert(3, 0)

        return camera_matrix



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import Poly3d

# == Fixtures =========================

@pytest.fixture
def poly3d_dict(
    sensor_lidar,
    attributes_multiple_types_dict,
    point3d_dict, point3d_another_dict
) -> dict:
    return {
        "uid": "9a9a30f5-D334-4f11-aa3f-c3c83f2935eb",
        "name": "lidar__poly3d__person",
        "val": point3d_dict + point3d_another_dict,
        "coordinate_system": sensor_lidar.uid,
        "attributes": attributes_multiple_types_dict,
        "closed": True,
    }

@pytest.fixture
def poly3d(
    point3d, point3d_another,
    sensor_lidar,
    object_person,
    attributes_multiple_types
) -> dict:
    return Poly3d(
        uid="9a9a30f5-D334-4f11-aa3f-c3c83f2935eb",
        points=[point3d, point3d_another],
        object=object_person,
        sensor=sensor_lidar,
        attributes=attributes_multiple_types,
        closed=True,
    )

# == Tests ============================

def test_fromdict(
    point3d, point3d_dict,
    point3d_another, point3d_another_dict,
    sensor_lidar, sensors,
    object_person,
    attributes_multiple_types, attributes_multiple_types_dict,
):
    poly3d = Poly3d.fromdict(
        {
            "uid": "9a9a30f5-D334-4f11-aa3f-c3c83f2935eb",
            "name": "lidar__poly3d__person",
            "val": point3d_dict + point3d_another_dict,
            "coordinate_system": sensor_lidar.uid,
            "attributes": attributes_multiple_types_dict,
            "closed": True,
        },
        sensors,
        object_person
    )

    assert poly3d.uid == "9a9a30f5-D334-4f11-aa3f-c3c83f2935eb"
    assert poly3d.name == "lidar__poly3d__person"
    assert poly3d.points == [point3d, point3d_another]
    assert poly3d.object == object_person
    assert poly3d.sensor == sensor_lidar
    assert poly3d.attributes == attributes_multiple_types
    assert poly3d.closed == True


def test_asdict(
    point3d, point3d_dict,
    point3d_another, point3d_another_dict,
    sensor_lidar,
    object_person,
    attributes_multiple_types, attributes_multiple_types_dict,
):
    poly3d = Poly3d(
        uid="9a9a30f5-D334-4f11-aa3f-c3c83f2935eb",
        points=[point3d, point3d_another],
        object=object_person,
        sensor=sensor_lidar,
        attributes=attributes_multiple_types,
        closed=True,
    )

    assert poly3d.asdict() == {
        "uid": "9a9a30f5-D334-4f11-aa3f-c3c83f2935eb",
        "name": "lidar__poly3d__person",
        "val": point3d_dict + point3d_another_dict,
        "coordinate_system": sensor_lidar.uid,
        "attributes": attributes_multiple_types_dict,
        "closed": True,
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

import raillabel


@pytest.fixture
def loader():
    return raillabel.load_.loader_classes.LoaderUnderstandAi()


def test_supports_true(json_data, loader):
    assert loader.supports(json_data["understand_ai_real_life"])


def test_supports_false(json_data, loader):
    data = json_data["understand_ai_real_life"]
    del data["metadata"]["project_id"]
    assert not loader.supports(data)


def test_load(json_data, loader):
    input_data_raillabel = remove_non_parsed_fields(json_data["openlabel_v1_short"])
    input_data_uai = json_data["understand_ai_t4_short"]

    scene_ground_truth = raillabel.load_.loader_classes.LoaderRailLabel().load(input_data_raillabel, validate=False)
    scene = loader.load(input_data_uai, validate=False)

    scene.metadata = scene_ground_truth.metadata

    assert scene.asdict() == scene_ground_truth.asdict()

def remove_non_parsed_fields(raillabel_data: dict) -> dict:
    """Return RailLabel file with frame_data and poly3ds removed."""

    for frame in raillabel_data["openlabel"]["frames"].values():

        if "frame_data" in frame["frame_properties"]:
            del frame["frame_properties"]["frame_data"]

        for object_id, object in list(frame["objects"].items()):
            if "poly3d" not in object["object_data"]:
                continue

            del object["object_data"]["poly3d"]
            if len(object["object_data"]) == 0:
                del frame["objects"][object_id]

    return raillabel_data


def test_raillabel_loader_warnings(loader):
    scene_dict = {
        "metadata": {
            "clip_id": "db_3_2021-09-22-14-28-01_2021-09-22-14-44-03",
            "external_clip_id": "2021-09-22-14-28-01_2021-09-22-14-44-03",
            "project_id": "trains_4",
            "export_time": "2023-04-20 01:38 UTC",
            "exporter_version": "1.0.0",
            "coordinate_system_3d": "FLU",
            "coordinate_system_reference": "SENSOR",
            "folder_name": "2021-09-22-14-28-01_2021-09-22-14-44-03"
        },
        "coordinateSystems": [],
        "frames": [
            {
                "frameId": "000",
                "timestamp": "1632321743.134149",
                "annotations": {
                    "2D_BOUNDING_BOX": [
                        {
                            "id": "78f0ad89-2750-4a30-9d66-44c9da73a714",
                            "objectId": "b40ba3ad-0327-46ff-9c28-2506cfd6d934",
                            "className": "2D_person",
                            "geometry": {
                                "xMin": -1.0,
                                "yMin": -0.5,
                                "xMax": 1,
                                "yMax": 2.5
                            },
                            "attributes": {},
                            "sensor": {
                                "type": "NON_EXISTENT_SENSOR",  # <-- relevant line
                                "uri": "S1206063/rgb_test0.png",
                                "timestamp": "1632321743.100000072"
                            }
                        }
                    ],
                    "2D_POLYLINE": [],
                    "2D_POLYGON": [],
                    "3D_BOUNDING_BOX": [],
                    "3D_SEGMENTATION": []
                }
            }
        ]
    }

    loader.load(scene_dict, validate=True)

    assert len(loader.warnings) == 2

    assert "NON_EXISTENT_SENSOR" in loader.warnings[0]
    assert "frame 0" in loader.warnings[0]

    assert "NON_EXISTENT_SENSOR" in loader.warnings[1]
    assert "78f0ad89-2750-4a30-9d66-44c9da73a714" in loader.warnings[1]

# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass

from ..._util._warning import _warning
from .sensor import Sensor


@dataclass
class Num:
    """A number.

    Parameters
    ----------
    uid: str
        This a string representing the unique universal identifier of the annotation.
    name: str
        Human readable name describing the annotation.
    val: int or float
        This is the value of the number object.
    attributes: dict, optional
        Attributes of the annotation. Dict keys are the uid str of the attribute, values are the
        attribute values. Default is {}.
    sensor: raillabel.format.Sensor, optional
        A reference to the sensor, this value is represented in. Default is None.
    """

    uid: str
    name: str
    val: t.Union[int, float]
    sensor: Sensor = None

    @classmethod
    def fromdict(cls, data_dict: dict, sensors: dict) -> "Num":
        """Generate a Num object from a dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        sensors: dict
            Dictionary containing all sensors for the scene.

        Returns
        -------
        annotation: Num
            Converted annotation.
        """

        return Num(
            uid=str(data_dict["uid"]),
            name=str(data_dict["name"]),
            val=data_dict["val"],
            sensor=cls._coordinate_system_fromdict(data_dict, sensors),
        )

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        return {
            "uid": str(self.uid),
            "name": str(self.name),
            "val": self.val,
            "coordinate_system": str(self.sensor.uid),
        }

    @classmethod
    def _coordinate_system_fromdict(cls, data_dict: dict, sensors: dict) -> t.Optional[Sensor]:

        is_coordinate_system_in_data = (
            "coordinate_system" in data_dict and data_dict["coordinate_system"] != ""
        )

        if not is_coordinate_system_in_data:
            return None

        if data_dict["coordinate_system"] not in sensors:
            _warning(
                f"'{data_dict['coordinate_system']}' does not exist as a sensor, "
                + f"but is referenced for the annotation {data_dict['uid']}."
            )
            return None

        return sensors[data_dict["coordinate_system"]]



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from abc import ABC, abstractmethod, abstractproperty

from ...format import Frame, _ObjectAnnotation


class _FilterABC(ABC):
    """ABC for all filter classes.

    Creating a new filter
    ---------------------
    To create a new, custom filter create a new class in this dir, that inherits from _FilterABC.
    Any class, that inherits from _FilterABC will automatically be loaded by the filter function.
    Include the filter arguments (include_[...], exclude_[...], ...) in the PARAMETERS field. These
    will be mutually exclusive.Select a level for the filter. The level determines where the filter
    is going to be applied (e.g. at the frame level, annotation level, ...). Include the conditions
    to pass the filter in the passes_filter() method, which returns True if the filter is passed.
    The contents of the filter arguments can optionally be processed by the _process_filter_args().
    """

    @property
    @abstractproperty
    def PARAMETERS(self) -> t.List[str]:
        raise NotImplementedError

    @property
    @abstractproperty
    def LEVELS(self) -> t.List[str]:
        raise NotImplementedError

    def __init__(self, kwargs):

        set_parameter = None
        for param in self.PARAMETERS:
            if param in kwargs and param is not None:

                if set_parameter is None:
                    setattr(self, param, self._process_filter_args(kwargs[param]))
                    set_parameter = param
                else:
                    raise ValueError(
                        f"{set_parameter} and {param} are mutually exclusive, but were both set."
                    )

            else:
                setattr(self, param, None)

    @abstractmethod
    def passes_filter(self, annotation: t.Union[t.Type[_ObjectAnnotation], Frame]) -> bool:
        raise NotImplementedError

    def _process_filter_args(self, filter_args):
        """Process filter arguments (optional)."""
        return filter_args



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel._util._warning import _WarningsLogger
from raillabel.format import Bbox

# == Fixtures =========================

@pytest.fixture
def bbox_dict(
    sensor_camera,
    attributes_multiple_types_dict,
    point2d_dict,
    size2d_dict,
) -> dict:
    return {
        "uid": "78f0ad89-2750-4a30-9d66-44c9da73a714",
        "name": "rgb_middle__bbox__person",
        "val": point2d_dict + size2d_dict,
        "coordinate_system": sensor_camera.uid,
        "attributes": attributes_multiple_types_dict
    }

@pytest.fixture
def bbox(
    point2d,
    size2d,
    sensor_camera,
    attributes_multiple_types,
    object_person,
) -> dict:
    return Bbox(
        uid="78f0ad89-2750-4a30-9d66-44c9da73a714",
        pos=point2d,
        size=size2d,
        sensor=sensor_camera,
        attributes=attributes_multiple_types,
        object=object_person,
    )


@pytest.fixture
def bbox_train_dict(sensor_camera, attributes_single_type_dict, point2d_dict, size2d_dict) -> dict:
    return {
        "uid": "6a7cfdb7-149d-4987-98dd-79d05a8cc8e6",
        "name": "rgb_middle__bbox__train",
        "val": point2d_dict + size2d_dict,
        "coordinate_system": sensor_camera.uid,
        "attributes": attributes_single_type_dict
    }

@pytest.fixture
def bbox_train(
    point2d,
    size2d,
    sensor_camera,
    attributes_single_type,
    object_train,
) -> dict:
    return Bbox(
        uid="6a7cfdb7-149d-4987-98dd-79d05a8cc8e6",
        pos=point2d,
        size=size2d,
        sensor=sensor_camera,
        attributes=attributes_single_type,
        object=object_train,
    )

# == Tests ============================

def test_fromdict(
    point2d, point2d_dict,
    size2d, size2d_dict,
    sensor_camera, sensors,
    object_person,
    attributes_multiple_types, attributes_multiple_types_dict,
):
    bbox = Bbox.fromdict(
        {
            "uid": "78f0ad89-2750-4a30-9d66-44c9da73a714",
            "name": "rgb_middle__bbox__person",
            "val": point2d_dict + size2d_dict,
            "coordinate_system": sensor_camera.uid,
            "attributes": attributes_multiple_types_dict
        },
        sensors,
        object_person
    )

    assert bbox.uid == "78f0ad89-2750-4a30-9d66-44c9da73a714"
    assert bbox.name == "rgb_middle__bbox__person"
    assert bbox.pos == point2d
    assert bbox.size == size2d
    assert bbox.object == object_person
    assert bbox.sensor == sensor_camera
    assert bbox.attributes == attributes_multiple_types

def test_fromdict_unknown_coordinate_system_warning(
    point2d_dict,
    size2d_dict,
    sensors,
    object_person,
):
    with _WarningsLogger() as logger:
        bbox = Bbox.fromdict(
            {
                "uid": "78f0ad89-2750-4a30-9d66-44c9da73a714",
                "name": "rgb_middle__bbox__person",
                "val": point2d_dict + size2d_dict,
                "coordinate_system": "UNKNOWN_COORDINATE_SYSTEM",
            },
            sensors,
            object_person
        )

    assert len(logger.warnings) == 1
    assert "78f0ad89-2750-4a30-9d66-44c9da73a714" in logger.warnings[0]
    assert "'UNKNOWN_COORDINATE_SYSTEM'" in logger.warnings[0]
    assert "sensor" in logger.warnings[0]

    assert bbox.sensor is None


def test_asdict(
    point2d, point2d_dict,
    size2d, size2d_dict,
    sensor_camera,
    object_person,
    attributes_multiple_types, attributes_multiple_types_dict,
):
    bbox = Bbox(
        uid="78f0ad89-2750-4a30-9d66-44c9da73a714",
        pos=point2d,
        size=size2d,
        object=object_person,
        sensor=sensor_camera,
        attributes=attributes_multiple_types,
    )

    assert bbox.asdict() == {
        "uid": "78f0ad89-2750-4a30-9d66-44c9da73a714",
        "name": "rgb_middle__bbox__person",
        "val": point2d_dict + size2d_dict,
        "coordinate_system": sensor_camera.uid,
        "attributes": attributes_multiple_types_dict
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass
from importlib import metadata as importlib_metadata


@dataclass
class Metadata:
    """Container for metadata information about the scene itself.

    As the OpenLABEL metadata object accepts additional properties, so does this class. Any
    properties present in the JSON will be added to the Metadata() object when read through
    Metadata.fromdict(). Conversely, all attributes from the Metadata() object will be stored
    into the JSON when using Metadata.asdict(). You can therefore just add attributes to the
    Python object and have them stored.
    Example:
        m = Metadata.fromdict(
            {
                "schema_version": "1.0.0",
                "some_additional_property": "Some Value"
            }
        )
        m.another_additional_property = "Another Value"
        m.asdict()
        -> {
            "schema_version": "1.0.0",
            "some_additional_property": "Some Value",
            "another_additional_property": "Another Value"
        }

    Parameters
    ----------
    schema_version: str
        Version number of the OpenLABEL schema this annotation object follows.
    annotator: str, optional
        Name or description of the annotator that created the annotations. Default is None.
    comment: str, optional
        Additional information or description about the annotation content. Default is None.
    exporter_version: str, optional
        Version of the raillabel-devkit, that last exported the scene. Default is None.
    file_version: str, optional
        Version number of the raillabel annotation content. Default is None.
    name: str, optional
        Name of the raillabel annotation content. Default is None.
    subschema_version: str, optional
        Version number of the RailLabel schema this annotation object follows. Default is None.
    tagged_file: str, optional
        Directory with the exported data_dict (e.g. images, point clouds). Default is None.
    """

    schema_version: str
    annotator: t.Optional[str] = None
    comment: t.Optional[str] = None
    exporter_version: t.Optional[str] = None
    file_version: t.Optional[str] = None
    name: t.Optional[str] = None
    subschema_version: t.Optional[str] = None
    tagged_file: t.Optional[str] = None

    @classmethod
    def fromdict(cls, data_dict: dict, subschema_version: t.Optional[str] = None) -> "Metadata":
        """Generate a Metadata object from a dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data. Additional (non-defined)
            arguments can be set and will be added as properties to Metadata.
        subschema_version: str, optional
            Version of the RailLabel subschema

        Returns
        -------
        metadata: Metadata
            Converted metadata.
        """

        metadata = Metadata(
            schema_version=data_dict["schema_version"],
            subschema_version=subschema_version,
            exporter_version=cls._collect_exporter_version(),
        )

        return cls._set_additional_attributes(metadata, data_dict)

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.
        """

        return self._remove_empty_fields(vars(self))

    @classmethod
    def _collect_exporter_version(cls) -> t.Optional[str]:

        try:
            exporter_version = importlib_metadata.version("raillabel")
        except importlib_metadata.PackageNotFoundError:
            return None

        version_number_length = len(exporter_version) - len(exporter_version.split(".")[-1])
        return exporter_version[: version_number_length - 1]

    @classmethod
    def _set_additional_attributes(cls, metadata: "Metadata", data_dict: dict) -> "Metadata":

        PRESET_KEYS = ["schema_version", "subschema_version", "exporter_version"]

        for key, value in data_dict.items():
            if key in PRESET_KEYS:
                continue

            is_key_a_valid_python_attribute = isinstance(key, str) and key.isidentifier()

            if not is_key_a_valid_python_attribute:
                raise KeyError(f"'{key}' is not a valid python attribute")

            setattr(metadata, key, value)

        return metadata

    def _remove_empty_fields(self, dict_repr: dict) -> dict:
        """Remove empty fields from a dictionary."""
        return {k: v for k, v in dict_repr.items() if v is not None}



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

# == Fixtures =========================

@pytest.fixture
def object_data_person_dict(bbox_dict, poly2d_dict, cuboid_dict, poly3d_dict, seg3d_dict) -> dict:
    return {
        "object_data": {
            "bbox": [bbox_dict],
            "poly2d": [poly2d_dict],
            "cuboid": [cuboid_dict],
            "poly3d": [poly3d_dict],
            "vec": [seg3d_dict],
        }
    }

@pytest.fixture
def object_data_train_dict(bbox_train_dict) -> dict:
    return {
        "object_data": {
            "bbox": [bbox_train_dict],
        }
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import FrameInterval

# == Fixtures =========================

@pytest.fixture
def frame_interval_dict() -> dict:
    return {
        "frame_start": 12,
        "frame_end": 16
    }

@pytest.fixture
def frame_interval() -> dict:
    return FrameInterval(
        frame_start=12,
        frame_end=16,
    )

# == Tests ============================

def test_fromdict():
    frame_interval = FrameInterval.fromdict(
        {
            "frame_start": 12,
            "frame_end": 16,
        }
    )

    assert frame_interval.frame_start == 12
    assert frame_interval.frame_end == 16


def test_asdict():
    frame_interval = FrameInterval(
        frame_start=12,
        frame_end=16,
    )

    assert frame_interval.asdict() == {
        "frame_start": 12,
        "frame_end": 16,
    }


def test_len():
    frame_interval = FrameInterval(
        frame_start=12,
        frame_end=16,
    )

    assert len(frame_interval) == 5


def test_from_frame_uids_empty():
    frame_uids = []

    assert FrameInterval.from_frame_uids(frame_uids) == []

def test_from_frame_uids_one_frame():
    frame_uids = [1]

    assert FrameInterval.from_frame_uids(frame_uids) == [
        FrameInterval(1, 1)
    ]

def test_from_frame_uids_one_interval():
    frame_uids = [1, 2, 3, 4]

    assert FrameInterval.from_frame_uids(frame_uids) == [
        FrameInterval(1, 4)
    ]

def test_from_frame_uids_multiple_intervals():
    frame_uids = [0, 1, 2, 3, 6, 7, 9, 12, 13, 14]

    assert FrameInterval.from_frame_uids(frame_uids) == [
        FrameInterval(0, 3),
        FrameInterval(6, 7),
        FrameInterval(9, 9),
        FrameInterval(12, 14),
    ]

def test_from_frame_uids_unsorted():
    frame_uids = [5, 2, 1, 3]

    assert FrameInterval.from_frame_uids(frame_uids) == [
        FrameInterval(1, 3),
        FrameInterval(5, 5),
    ]


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import Num

# == Fixtures =========================

@pytest.fixture
def num_dict(sensor_camera) -> dict:
    return {
        "uid": "4e86c449-3B19-410c-aa64-603d46da3b26",
        "name": "some_number",
        "val": 24,
        "coordinate_system": sensor_camera.uid
    }

@pytest.fixture
def num(sensor_camera) -> dict:
    return Num(
        uid="4e86c449-3B19-410c-aa64-603d46da3b26",
        name="some_number",
        val=24,
        sensor=sensor_camera,
    )

# == Tests ============================

def test_fromdict(sensor_camera):
    num = Num.fromdict(
        {
            "uid": "4e86c449-3B19-410c-aa64-603d46da3b26",
            "name": "some_number",
            "val": 24,
            "coordinate_system": sensor_camera.uid,
        },
        {
            sensor_camera.uid: sensor_camera
        }
    )

    assert num.uid == "4e86c449-3B19-410c-aa64-603d46da3b26"
    assert num.name == "some_number"
    assert num.val == 24
    assert num.sensor == sensor_camera


def test_asdict(sensor_camera):
    num = Num(
        uid="4e86c449-3B19-410c-aa64-603d46da3b26",
        name="some_number",
        val=24,
        sensor=sensor_camera,
    )

    assert num.asdict() == {
        "uid": "4e86c449-3B19-410c-aa64-603d46da3b26",
        "name": "some_number",
        "val": 24,
        "coordinate_system": sensor_camera.uid,
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

import raillabel.format.understand_ai as uai_format

# == Fixtures =========================

@pytest.fixture
def coordinate_system_camera_uai_dict(point_3d_vec, quaternion_vec) -> dict:
    return {
        "coordinate_system_id": "ir_middle",
        "topic": "/A0001781/image",
        "frame_id": " A0001781",
        "position": point_3d_vec,
        "rotation_quaternion": quaternion_vec,
        "rotation_matrix": [0] * 9,
        "angle_axis_rotation": [0] * 3,
        "homogeneous_transform": [0] * 16,
        "measured_position": [0, 0, 0],
        "camera_matrix": [
            3535,    0, 319.5,
               0, 3535, 239.5,
               0,    0,   1  ,
        ],
        "dist_coeffs": [0, 1, 2, 3, 4]
    }

@pytest.fixture
def coordinate_system_camera_uai(point_3d_vec, quaternion_vec):
    return uai_format.CoordinateSystem(
        uid="ir_middle",
        topic="/A0001781/image",
        frame_id=" A0001781",
        position=point_3d_vec,
        rotation_quaternion=quaternion_vec,
        rotation_matrix=[0] * 9,
        angle_axis_rotation=[0] * 3,
        homogeneous_transform=[0] * 16,
        measured_position=[0, 0, 0],
        camera_matrix=[
            3535,    0, 319.5,
               0, 3535, 239.5,
               0,    0,   1  ,
        ],
        dist_coeffs=[0, 1, 2, 3, 4]
    )

@pytest.fixture
def coordinate_system_camera_raillabel_dict(point_3d_vec, quaternion_vec) -> dict:
    return (
        {
            "type": "sensor",
            "parent": "base",
            "pose_wrt_parent": {
                "translation": point_3d_vec,
                "quaternion": quaternion_vec
            }
        },
        {
            "type": "camera",
            "uri": "/A0001781/image",
            "stream_properties": {
                "intrinsics_pinhole": {
                    "camera_matrix": [
                        3535,    0, 319.5, 0,
                        0, 3535, 239.5, 0,
                        0,    0,   1  , 0,
                    ],
                    "distortion_coeffs": [0, 1, 2, 3, 4],
                    "width_px": 640,
                    "height_px": 480,
                }
            }
        }
    )

@pytest.fixture
def coordinate_system_camera_translated_uid() -> dict:
    return "ir_middle"


@pytest.fixture
def coordinate_system_lidar_uai_dict(point_3d_vec, quaternion_vec) -> dict:
    return {
        "coordinate_system_id": "LIDAR",
        "topic": "/lidar_merged",
        "frame_id": "lidar_merged",
        "position": point_3d_vec,
        "rotation_quaternion": quaternion_vec,
        "rotation_matrix": [0] * 9,
        "angle_axis_rotation": [0] * 3,
        "homogeneous_transform": [0] * 16,
    }

@pytest.fixture
def coordinate_system_lidar_uai(point_3d_vec, quaternion_vec):
    return uai_format.CoordinateSystem(
        uid="LIDAR",
        topic="/lidar_merged",
        frame_id="lidar_merged",
        position=point_3d_vec,
        rotation_quaternion=quaternion_vec,
        rotation_matrix=[0] * 9,
        angle_axis_rotation=[0] * 3,
        homogeneous_transform=[0] * 16,
    )

@pytest.fixture
def coordinate_system_lidar_raillabel_dict(point_3d_vec, quaternion_vec) -> dict:
    return (
        {
            "type": "sensor",
            "parent": "base",
            "pose_wrt_parent": {
                "translation": point_3d_vec,
                "quaternion": quaternion_vec
            }
        },
        {
            "type": "lidar",
            "uri": "/lidar_merged",
        }
    )

@pytest.fixture
def coordinate_system_lidar_translated_uid() -> dict:
    return "lidar"


@pytest.fixture
def coordinate_system_radar_uai_dict(point_3d_vec, quaternion_vec) -> dict:
    return {
        "coordinate_system_id": "radar",
        "rotation_around_z_in_degrees": 1.22869,
        "topic": "/talker1/Nvt/Cartesian",
        "frame_id": "navtech",
        "position": point_3d_vec,
        "rotation_quaternion": quaternion_vec,
        "rotation_matrix": [0] * 9,
        "angle_axis_rotation": [0] * 3,
        "homogeneous_transform": [0] * 16,
    }

@pytest.fixture
def coordinate_system_radar_uai(point_3d_vec, quaternion_vec):
    return uai_format.CoordinateSystem(
        uid="radar",
        topic="/talker1/Nvt/Cartesian",
        frame_id="navtech",
        position=point_3d_vec,
        rotation_quaternion=quaternion_vec,
        rotation_matrix=[0] * 9,
        angle_axis_rotation=[0] * 3,
        homogeneous_transform=[0] * 16,
    )

@pytest.fixture
def coordinate_system_radar_raillabel_dict(point_3d_vec, quaternion_vec) -> dict:
    return (
        {
            "type": "sensor",
            "parent": "base",
            "pose_wrt_parent": {
                "translation": point_3d_vec,
                "quaternion": quaternion_vec
            }
        },
        {
            "type": "radar",
            "uri": "/talker1/Nvt/Cartesian",
            "stream_properties": {
                "intrinsics_radar": {
                    "resolution_px_per_m": 2.856,
                    "width_px": 2856,
                    "height_px": 1428
                }
            }
        }
    )

@pytest.fixture
def coordinate_system_radar_translated_uid() -> dict:
    return "radar"

# == Tests ============================

def test_fromdict(point_3d_vec, quaternion_vec):
    coordinate_system = uai_format.CoordinateSystem.fromdict(
        {
            "coordinate_system_id": "ir_middle",
            "topic": "/A0001781/image",
            "frame_id": " A0001781",
            "position": point_3d_vec,
            "rotation_quaternion": quaternion_vec,
            "rotation_matrix": [0] * 9,
            "angle_axis_rotation": [0] * 3,
            "homogeneous_transform": [0] * 16,
            "measured_position": [0, 0, 0],
            "camera_matrix": [
                3535,    0, 319.5,
                   0, 3535, 239.5,
                   0,    0,   1  ,
            ],
            "dist_coeffs": [0, 1, 2, 3, 4]
        }
    )

    assert coordinate_system.uid == "ir_middle"
    assert coordinate_system.topic == "/A0001781/image"
    assert coordinate_system.frame_id == " A0001781"
    assert coordinate_system.position == point_3d_vec
    assert coordinate_system.rotation_quaternion == quaternion_vec
    assert coordinate_system.rotation_matrix == [0] * 9
    assert coordinate_system.angle_axis_rotation == [0] * 3
    assert coordinate_system.homogeneous_transform == [0] * 16
    assert coordinate_system.measured_position == [0, 0, 0]
    assert coordinate_system.camera_matrix == [
        3535,    0, 319.5,
           0, 3535, 239.5,
           0,    0,   1  ,
    ]
    assert coordinate_system.dist_coeffs == [0, 1, 2, 3, 4]


def test_to_raillabel__coordinate_system(point_3d_vec, quaternion_vec):
    coordinate_system = uai_format.CoordinateSystem(
        uid="ir_middle",
        topic="/A0001781/image",
        frame_id=" A0001781",
        position=point_3d_vec,
        rotation_quaternion=quaternion_vec,
        rotation_matrix=[0] * 9,
        angle_axis_rotation=[0] * 3,
        homogeneous_transform=[0] * 16,
        measured_position=[0, 0, 0],
        camera_matrix=[
            3535,    0, 319.5,
               0, 3535, 239.5,
               0,    0,   1  ,
        ],
        dist_coeffs=[0, 1, 2, 3, 4]
    )

    assert coordinate_system.to_raillabel()[0] == {
        "type": "sensor",
        "parent": "base",
        "pose_wrt_parent": {
            "translation": point_3d_vec,
            "quaternion": quaternion_vec
        }
    }

def test_to_raillabel__stream__camera(point_3d_vec, quaternion_vec):
    coordinate_system = uai_format.CoordinateSystem(
        uid="ir_middle",
        topic="/A0001781/image",
        frame_id=" A0001781",
        position=point_3d_vec,
        rotation_quaternion=quaternion_vec,
        rotation_matrix=[0] * 9,
        angle_axis_rotation=[0] * 3,
        homogeneous_transform=[0] * 16,
        measured_position=[0, 0, 0],
        camera_matrix=[
            3535,    0, 319.5,
               0, 3535, 239.5,
               0,    0,   1  ,
        ],
        dist_coeffs=[0, 1, 2, 3, 4]
    )

    assert coordinate_system.to_raillabel()[1] == {
        "type": "camera",
        "uri": "/A0001781/image",
        "stream_properties": {
            "intrinsics_pinhole": {
                "camera_matrix": [
                    3535,    0, 319.5, 0,
                       0, 3535, 239.5, 0,
                       0,    0,   1  , 0,
                ],
                "distortion_coeffs": [0, 1, 2, 3, 4],
                "width_px": 640,
                "height_px": 480,
            }
        }
    }

def test_to_raillabel__stream__lidar(point_3d_vec, quaternion_vec):
    coordinate_system = uai_format.CoordinateSystem(
        uid="LIDAR",
        topic="/lidar_merged",
        frame_id="lidar_merged",
        position=point_3d_vec,
        rotation_quaternion=quaternion_vec,
        rotation_matrix=[0] * 9,
        angle_axis_rotation=[0] * 3,
        homogeneous_transform=[0] * 16,
    )

    assert coordinate_system.to_raillabel()[1] == {
        "type": "lidar",
        "uri": "/lidar_merged",
    }

def test_to_raillabel__stream__radar(point_3d_vec, quaternion_vec):
    coordinate_system = uai_format.CoordinateSystem(
        uid="radar",
        topic="/talker1/Nvt/Cartesian",
        frame_id="navtech",
        position=point_3d_vec,
        rotation_quaternion=quaternion_vec,
        rotation_matrix=[0] * 9,
        angle_axis_rotation=[0] * 3,
        homogeneous_transform=[0] * 16,
    )

    assert coordinate_system.to_raillabel()[1] == {
        "type": "radar",
        "uri": "/talker1/Nvt/Cartesian",
        "stream_properties": {
            "intrinsics_radar": {
                "resolution_px_per_m": 2.856,
                "width_px": 2856,
                "height_px": 1428
            }
        }
    }


if __name__ == "__main__":
    import os
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import _ObjectAnnotation

# == Fixtures =========================

@pytest.fixture
def attributes_single_type_dict() -> dict:
    return {
        "text": [
            {
                "name": "test_text_attr0",
                "val": "test_text_attr0_val"
            },
            {
                "name": "test_text_attr1",
                "val": "test_text_attr1_val"
            }
        ]
    }

@pytest.fixture
def attributes_single_type() -> dict:
    return {
        "test_text_attr0": "test_text_attr0_val",
        "test_text_attr1": "test_text_attr1_val",
    }


@pytest.fixture
def attributes_multiple_types_dict() -> dict:
    return {
        "text": [{
            "name": "text_attr",
            "val": "text_val"
        }],
        "num": [{
            "name": "num_attr",
            "val": 0
        }],
        "boolean": [{
            "name": "bool_attr",
            "val": True
        }],
        "vec": [{
            "name": "vec_attr",
            "val": [0, 1, 2]
        }],
    }

@pytest.fixture
def attributes_multiple_types() -> dict:
    return {
        "text_attr": "text_val",
        "num_attr": 0,
        "bool_attr": True,
        "vec_attr": [0, 1, 2],
    }

# == Tests ============================

def test_fromdict__single_type():
    attributes_dict = {
        "attributes": {
            "text": [
                {
                    "name": "test_text_attr0",
                    "val": "test_text_attr0_val"
                },
                {
                    "name": "test_text_attr1",
                    "val": "test_text_attr1_val"
                }
            ]
        }
    }

    assert _ObjectAnnotation._attributes_fromdict(attributes_dict) == {
        "test_text_attr0": "test_text_attr0_val",
        "test_text_attr1": "test_text_attr1_val",
    }

def test_fromdict__multiple_types():
    attributes_dict = {
        "attributes": {
            "text": [{
                "name": "text_attr",
                "val": "text_val"
            }],
            "num": [{
                "name": "num_attr",
                "val": 0
            }],
            "boolean": [{
                "name": "bool_attr",
                "val": True
            }],
            "vec": [{
                "name": "vec_attr",
                "val": [0, 1, 2]
            }],
        }
    }

    assert _ObjectAnnotation._attributes_fromdict(attributes_dict) == {
        "text_attr": "text_val",
        "num_attr": 0,
        "bool_attr": True,
        "vec_attr": [0, 1, 2],
    }


def test_asdict__single_type():
    attributes = {
        "test_text_attr0": "test_text_attr0_val",
        "test_text_attr1": "test_text_attr1_val",
    }

    assert _ObjectAnnotation._attributes_asdict(None, attributes) == {
        "text": [
            {
                "name": "test_text_attr0",
                "val": "test_text_attr0_val"
            },
            {
                "name": "test_text_attr1",
                "val": "test_text_attr1_val"
            }
        ]
    }

def test_asdict__multiple_types():
    attributes = {
        "text_attr": "text_val",
        "num_attr": 0,
        "bool_attr": True,
        "vec_attr": [0, 1, 2],
    }

    assert _ObjectAnnotation._attributes_asdict(None, attributes) == {
        "text": [{
            "name": "text_attr",
            "val": "text_val"
        }],
        "num": [{
            "name": "num_attr",
            "val": 0
        }],
        "boolean": [{
            "name": "bool_attr",
            "val": True
        }],
        "vec": [{
            "name": "vec_attr",
            "val": [0, 1, 2]
        }],
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass

from .point3d import Point3d
from .quaternion import Quaternion


@dataclass
class Transform:
    """A transformation between two coordinate systems.

    Parameters
    ----------
    pos: raillabel.format.Point3d
        Translation with regards to the parent coordinate system.
    quat: raillabel.format.Quaternion
        Rotation quaternion with regards to the parent coordinate system.
    """

    pos: Point3d
    quat: Quaternion

    @classmethod
    def fromdict(cls, data_dict: dict) -> "Transform":
        """Generate a Transform object from a dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        """

        return Transform(
            pos=Point3d.fromdict(data_dict["translation"]),
            quat=Quaternion.fromdict(data_dict["quaternion"]),
        )

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = {
            "translation": self.pos.asdict(),
            "quaternion": self.quat.asdict(),
        }

        return dict_repr



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass


@dataclass
class IntrinsicsPinhole:
    """Intrinsic calibration for a camera sensor.

    Parameters
    ----------
    camera_matrix: tuple of float of length 12
        This is a 3x4 camera matrix which projects 3D homogeneous points (4x1) from a camera
        coordinate system into the image plane (3x1). This is the usual K matrix for camera
        projection as in OpenCV. It is extended from 3x3 to 3x4 to enable its direct utilisation to
        project 4x1 homogeneous 3D points. The matrix is defined to follow the camera model:
        x-to-right, y-down, z-forward. The following equation applies:
        x_img = camera_matrix * X_ccs.
    distortion: tuple of float of length 5 to 14
        This is the array 1xN radial and tangential distortion coefficients.
    width_px: int
        Width of the image frame in pixels.
    height_px: int
        Height of the image frame in pixels.
    """

    camera_matrix: t.Tuple[float, ...]
    distortion: t.Tuple[float, ...]
    width_px: int
    height_px: int

    @classmethod
    def fromdict(cls, data_dict: dict) -> "IntrinsicsPinhole":
        """Generate a IntrinsicsPinhole object from a dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data.

        Returns
        -------
        raillabel.format.IntrinsicsPinhole
            Converted IntrinsicsPinhole object.
        """

        return IntrinsicsPinhole(
            camera_matrix=tuple(data_dict["camera_matrix"]),
            distortion=tuple(data_dict["distortion_coeffs"]),
            width_px=data_dict["width_px"],
            height_px=data_dict["height_px"],
        )

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = {
            "camera_matrix": list(self.camera_matrix),
            "distortion_coeffs": list(self.distortion),
            "width_px": int(self.width_px),
            "height_px": int(self.height_px),
        }

        return dict_repr



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import sys
from pathlib import Path
from uuid import UUID

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

import raillabel.format.understand_ai as uai_format
from raillabel.format.understand_ai._translation import translate_class_id

# == Fixtures =========================

@pytest.fixture
def segmentation_3d_uai_dict(sensor_lidar_uai_dict, attributes_uai_dict) -> dict:
    return {
        "id": "13478f94-d556-4f64-a72b-47662e94988e",
        "objectId": "05a7e7a7-91e1-49ef-a172-780f2461f013",
        "className": "test_class",
        "geometry": {
            "associatedPoints": [39814, 39815, 39816, 39817, 39818],
            "numberOfPointsInBox": 5
        },
        "attributes": attributes_uai_dict,
        "sensor": sensor_lidar_uai_dict,
    }

@pytest.fixture
def segmentation_3d_uai(attributes_uai, sensor_lidar_uai) -> dict:
    return uai_format.Segmentation3d(
        id=UUID("13478f94-d556-4f64-a72b-47662e94988e"),
        object_id=UUID("05a7e7a7-91e1-49ef-a172-780f2461f013"),
        class_name="test_class",
        associated_points=[39814, 39815, 39816, 39817, 39818],
        number_of_points=5,
        attributes=attributes_uai,
        sensor=sensor_lidar_uai,
    )

@pytest.fixture
def segmentation_3d_raillabel_dict(attributes_raillabel_dict, coordinate_system_lidar_translated_uid) -> dict:
    return {
        "name": "13478f94-d556-4f64-a72b-47662e94988e",
        "val": [39814, 39815, 39816, 39817, 39818],
        "coordinate_system": coordinate_system_lidar_translated_uid,
        "attributes": attributes_raillabel_dict
    }

# == Tests ============================

def test_fromdict(
    attributes_uai_dict, attributes_uai,
    sensor_lidar_uai_dict, sensor_lidar_uai
):
    segmentation_3d = uai_format.Segmentation3d.fromdict(
        {
            "id": "13478f94-d556-4f64-a72b-47662e94988e",
            "objectId": "05a7e7a7-91e1-49ef-a172-780f2461f013",
            "className": "test_class",
            "geometry": {
                "associatedPoints": [39814, 39815, 39816, 39817, 39818],
                "numberOfPointsInBox": 5
            },
            "attributes": attributes_uai_dict,
            "sensor": sensor_lidar_uai_dict,
        }
    )

    assert segmentation_3d.id == UUID("13478f94-d556-4f64-a72b-47662e94988e")
    assert segmentation_3d.object_id == UUID("05a7e7a7-91e1-49ef-a172-780f2461f013")
    assert segmentation_3d.class_name == "test_class"
    assert segmentation_3d.associated_points == [39814, 39815, 39816, 39817, 39818]
    assert segmentation_3d.number_of_points == 5
    assert segmentation_3d.attributes == attributes_uai
    assert segmentation_3d.sensor == sensor_lidar_uai


def test_to_raillabel(
    attributes_uai, attributes_raillabel_dict,
    sensor_lidar_uai, sensor_lidar_raillabel_dict, coordinate_system_lidar_translated_uid,
):
    segmentation_3d = uai_format.Segmentation3d(
        id=UUID("13478f94-d556-4f64-a72b-47662e94988e"),
        object_id=UUID("05a7e7a7-91e1-49ef-a172-780f2461f013"),
        class_name="test_class",
        associated_points=[39814, 39815, 39816, 39817, 39818],
        number_of_points=5,
        attributes=attributes_uai,
        sensor=sensor_lidar_uai,
    )

    data_dict, object_id, translated_class_id, sensor_reference = segmentation_3d.to_raillabel()

    assert data_dict == {
        "name": "13478f94-d556-4f64-a72b-47662e94988e",
        "val": [39814, 39815, 39816, 39817, 39818],
        "coordinate_system": coordinate_system_lidar_translated_uid,
        "attributes": attributes_raillabel_dict,
    }
    assert object_id == str(segmentation_3d.object_id)
    assert translated_class_id == translate_class_id(segmentation_3d.class_name)
    assert sensor_reference == sensor_lidar_raillabel_dict

if __name__ == "__main__":
    import os
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass
from decimal import Decimal


@dataclass
class SensorReference:
    """Information for a sensor in a frame.

    Parameters
    ----------
    type: str
        Friendly name of the sensor and its unique identifier.
    uri: str
        URI to the file containing the frame specific sensor output from the project directory.
    timestamp: decimal.Decimal
        Unix timestamp of the sensor recording.
    """

    type: str
    uri: str
    timestamp: Decimal

    @classmethod
    def fromdict(cls, data_dict: dict) -> "SensorReference":
        """Generate a SensorReference from a dictionary in the UAI format.

        Parameters
        ----------
        data_dict: dict
            Understand.AI T4 format dictionary containing the data_dict.

        Returns
        -------
        SensorReference
            Converted sensor reference.
        """

        return SensorReference(
            type=data_dict["type"], uri=data_dict["uri"], timestamp=Decimal(data_dict["timestamp"])
        )

    def to_raillabel(self) -> t.Tuple[str, dict]:
        """Convert to a raillabel compatible dict.

        Returns
        -------
        sensor_id: str
            Friendly identifier of the sensor.
        sensor_reference: dict
            Dictionary valid for the raillabel schema.
        """

        return (
            self.type,
            {
                "stream_properties": {"sync": {"timestamp": str(self.timestamp)}},
                "uri": self.uri.split("/")[-1],
            },
        )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Module containing all relevant understand.ai format classes."""

from .bounding_box_2d import BoundingBox2d
from .bounding_box_3d import BoundingBox3d
from .coordinate_system import CoordinateSystem
from .frame import Frame
from .metadata import Metadata
from .point_3d import Point3d
from .polygon_2d import Polygon2d
from .polyline_2d import Polyline2d
from .quaternion import Quaternion
from .scene import Scene
from .segmentation_3d import Segmentation3d
from .sensor_reference import SensorReference
from .size_3d import Size3d



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0


class UnsupportedFormatError(Exception):
    """Raised when a loaded annotation file is not in a supported format."""

    __module__ = "raillabel"


class SchemaError(Exception):
    """Raised when the data does not validate against a given schema."""

    __module__ = "raillabel"


class AmbiguousSchemaNameError(Exception):
    """Raised when a schema key applies to more than one schema files in.

    /schemas.
    """

    __module__ = "raillabel"


class MissingStreamError(Exception):
    """Raised when a coordinate system has no corresponding stream."""

    __module__ = "raillabel"


class MissingCoordinateSystemError(Exception):
    """Raised when a stream has no corresponding coordinate system."""

    __module__ = "raillabel"


class UnsupportedParentError(Exception):
    """Raised when a coordinate system does not have 'base' as parent."""

    __module__ = "raillabel"



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass

from ..._util._attribute_type import AttributeType
from .frame_interval import FrameInterval


@dataclass
class ElementDataPointer:
    """Container of pointers for annotations indexed by the "name" property.

    Used for indexing annotations for easy referencing without loading all of them.

    Parameters
    ----------
    uid: str
        Unique identifier of the ElementDataPointer assembled using a specific schema.
    frame_intervals: list[raillabel.format.FrameInterval]
        Frame intervals, this element data pointer is contained in.
    attribute_pointers: dict[str, raillabel.util._attribute_type.AttributeType]
        References of attributes contained in the referenced annotations with attribute type.

    Properties (read-only)
    ----------------------
    uid: str
        Unique identifier of the ElementDataPointer built from the attributes.
    """

    uid: str
    frame_intervals: t.List[FrameInterval]
    attribute_pointers: t.Dict[str, AttributeType]

    @property
    def annotation_type(self) -> str:
        """Return type of annotation e.g. bbox, cuboid."""
        return self.uid.split("__")[1]

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        return {
            "type": self.annotation_type,
            "frame_intervals": self._frame_intervals_asdict(),
            "attribute_pointers": self._attribute_pointers_asdict(),
        }

    def _frame_intervals_asdict(self) -> t.List[t.Dict[str, int]]:
        return [fi.asdict() for fi in self.frame_intervals]

    def _attribute_pointers_asdict(self) -> t.Dict[str, str]:
        return {
            attr_name: attr_type.value for attr_name, attr_type in self.attribute_pointers.items()
        }



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass, field

from ... import exceptions
from ..._util._clean_dict import _clean_dict
from ..._util._warning import _warning
from .frame import Frame
from .frame_interval import FrameInterval
from .metadata import Metadata
from .object import Object
from .sensor import Sensor


@dataclass
class Scene:
    """The root RailLabel class, which contains all data.

    Parameters
    ----------
    metadata: raillabel.format.Metadata
        This object contains information, that is, metadata, about the annotation file itself.
    sensors: dict of raillabel.format.Sensor, optional
        Dictionary of raillabel.format.Sensors. Dictionary keys are the sensor uids. Default is {}.
    objects: dict of raillabel.format.Object, optional
        Dictionary of raillabel.format.Objects. Dictionary keys are the object uids. Default is {}.
    frames: dict of raillabel.format.Frame, optional
        Dict of frames in the scene. Dictionary keys are the frame uids. Default is {}.

    Properties (read-only)
    ----------------------
    frame_intervals: list[FrameIntervals]
        List of frame intervals describing the frames present in this scene.
    """

    metadata: Metadata
    sensors: t.Dict[str, Sensor] = field(default_factory=dict)
    objects: t.Dict[str, Object] = field(default_factory=dict)
    frames: t.Dict[int, Frame] = field(default_factory=dict)

    @property
    def frame_intervals(self) -> t.List[FrameInterval]:
        """Return frame intervals of the present frames."""
        return FrameInterval.from_frame_uids(list(self.frames.keys()))

    # === Public Methods ==========================================================================

    @classmethod
    def fromdict(cls, data_dict: dict, subschema_version: t.Optional[str] = None) -> "Scene":
        """Generate a Scene object from a RailLABEL-dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        subschema_version: str, optional
            Version of the RailLabel subschema.

        Returns
        -------
        raillabel.format.Scene
            Converted Scene object.

        Raises
        ------
        raillabel.exceptions.MissingCoordinateSystemError
            if a stream has no corresponding coordinate system.
        raillabel.exceptions.MissingStreamError
            if a coordinate system has no corresponding stream.
        raillabel.exceptions.UnsupportedParentError
            if a coordinate system has no corresponding stream.
        """

        data_dict = cls._prepare_data(data_dict)

        sensors = cls._sensors_fromdict(data_dict["streams"], data_dict["coordinate_systems"])
        objects = cls._objects_fromdict(data_dict["objects"])

        return Scene(
            metadata=Metadata.fromdict(data_dict["metadata"], subschema_version),
            sensors=sensors,
            objects=objects,
            frames=cls._frames_fromdict(data_dict["frames"], sensors, objects),
        )

    def asdict(self, calculate_pointers: bool = True) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this Scene.
        calculate_pointers: bool, optional
            If True, object_data_pointers and Object frame_intervals will be calculated. Default
            is True.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        return {
            "openlabel": _clean_dict(
                {
                    "metadata": self.metadata.asdict(),
                    "streams": self._streams_asdict(self.sensors),
                    "coordinate_systems": self._coordinate_systems_asdict(self.sensors),
                    "objects": self._objects_asdict(self.objects, calculate_pointers),
                    "frames": self._frames_asdict(self.frames),
                    "frame_intervals": self._frame_intervals_asdict(self.frame_intervals),
                }
            )
        }

    # === Private Methods =========================================================================

    # --- fromdict() ----------------------------

    @classmethod
    def _prepare_data(cls, data: dict) -> dict:
        """Add optional fields to dict to simplify interaction.

        Parameters
        ----------
        data : dict
            JSON data.

        Returns
        -------
        dict
            Enhanced JSON data.
        """

        if "coordinate_systems" not in data["openlabel"]:
            data["openlabel"]["coordinate_systems"] = {}

        if "streams" not in data["openlabel"]:
            data["openlabel"]["streams"] = {}

        if "objects" not in data["openlabel"]:
            data["openlabel"]["objects"] = {}

        if "frames" not in data["openlabel"]:
            data["openlabel"]["frames"] = {}

        return data["openlabel"]

    @classmethod
    def _sensors_fromdict(
        cls, streams_dict: dict, coordinate_systems_dict: dict
    ) -> t.Dict[str, Sensor]:

        cls._check_sensor_completeness(streams_dict, coordinate_systems_dict)

        sensors = {}

        for stream_id in streams_dict:
            sensors[stream_id] = Sensor.fromdict(
                uid=stream_id,
                cs_data_dict=coordinate_systems_dict[stream_id],
                stream_data_dict=streams_dict[stream_id],
            )

        return sensors

    @classmethod
    def _check_sensor_completeness(cls, streams_dict: dict, coordinate_systems_dict: dict):

        for stream_uid in streams_dict:
            if stream_uid not in coordinate_systems_dict:
                raise exceptions.MissingCoordinateSystemError(
                    f"Stream {stream_uid} has no corresponding coordinate system."
                )

        for cs_uid in coordinate_systems_dict:
            if cs_uid == "base":
                continue

            if coordinate_systems_dict[cs_uid]["parent"] != "base":
                raise exceptions.UnsupportedParentError(
                    f"Only 'base' is permitted as a parent for coordinate system {cs_uid}, "
                    + f"not {coordinate_systems_dict[cs_uid]['parent']}."
                )

            if cs_uid not in streams_dict:
                raise exceptions.MissingStreamError(
                    f"Coordinate sytem {cs_uid} has no corresponding stream."
                )

    @classmethod
    def _objects_fromdict(cls, object_dict: dict) -> t.Dict[str, Object]:
        return {uid: Object.fromdict(object, uid) for uid, object in object_dict.items()}

    @classmethod
    def _frames_fromdict(
        cls, frames_dict: dict, sensors: t.Dict[str, Sensor], objects: t.Dict[str, Object]
    ) -> t.Dict[int, Frame]:

        frames = {}
        for frame_uid, frame_dict in frames_dict.items():
            frame_uid = int(frame_uid)

            if frame_uid in frames:
                _warning(
                    f"Frame UID {frame_uid} is contained more than once in the scene. "
                    + "The duplicate frame will be omitted."
                )
                continue

            frames[frame_uid] = Frame.fromdict(frame_uid, frame_dict, objects, sensors)

        return frames

    # --- asdict() ------------------------------

    def _streams_asdict(self, sensors: t.Dict[str, Sensor]) -> dict:
        return {uid: sensor.asdict()["stream"] for uid, sensor in sensors.items()}

    def _coordinate_systems_asdict(self, sensors: t.Dict[str, Sensor]) -> dict:

        if len(sensors) == 0:
            return None

        coordinate_systems = {"base": {"type": "local", "parent": "", "children": []}}

        for uid, sensor in sensors.items():
            coordinate_systems[uid] = sensor.asdict()["coordinate_system"]
            coordinate_systems["base"]["children"].append(uid)

        return coordinate_systems

    def _objects_asdict(self, objects: t.Dict[str, Object], calculate_pointers: bool) -> dict:

        if calculate_pointers:
            return {str(uid): object.asdict(self.frames) for uid, object in objects.items()}
        else:
            return {str(uid): object.asdict() for uid, object in objects.items()}

    def _frames_asdict(self, frames: t.Dict[int, Frame]) -> dict:
        return {str(uid): frame.asdict() for uid, frame in frames.items()}

    def _frame_intervals_asdict(self, frame_intervals: t.List[FrameInterval]) -> t.List[dict]:
        return [fi.asdict() for fi in frame_intervals]



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass


@dataclass
class FrameInterval:
    """Closed interval of frames.

    Parameters
    ----------
    frame_start: int
        Initial frame number of the interval (inclusive).
    frame_end: int
        Ending frame number of the interval (inclusive).
    """

    frame_start: int
    frame_end: int

    @classmethod
    def fromdict(cls, data_dict: dict) -> "FrameInterval":
        """Generate a FrameInterval object from a dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        """

        return FrameInterval(
            frame_start=data_dict["frame_start"],
            frame_end=data_dict["frame_end"],
        )

    @classmethod
    def from_frame_uids(cls, frame_uids: t.List[int]) -> t.List["FrameInterval"]:
        """Convert a list of frame uids into FrameIntervals.

        Parameters
        ----------
        frame_uids: list[int]
            List of frame uids, that should be included in the FrameIntervals.

        Returns
        -------
        list[FrameInterval]
            FrameIntervals corresponding to the frames ids.

        Example
        -------
        FrameInterval.from_frame_uids([0, 1, 2, 3, 6, 7, 9, 12, 13, 14]) == [
            FrameInterval(0, 3),
            FrameInterval(6, 7),
            FrameInterval(9, 9),
            FrameInterval(12, 14),
        ]
        """

        sorted_frame_uids = sorted(frame_uids)
        frame_uid_intervals = cls._slice_into_intervals(sorted_frame_uids)

        frame_intervals = []
        for interval in frame_uid_intervals:
            frame_intervals.append(FrameInterval(frame_start=interval[0], frame_end=interval[-1]))

        return frame_intervals

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        return {
            "frame_start": int(self.frame_start),
            "frame_end": int(self.frame_end),
        }

    def __len__(self) -> int:
        """Return the length in frames."""
        return abs(self.frame_start - self.frame_end) + 1

    @classmethod
    def _slice_into_intervals(cls, sorted_frame_uids: t.List[int]) -> t.List[t.List[int]]:

        if len(sorted_frame_uids) == 0:
            return []

        if len(sorted_frame_uids) == 1:
            return [sorted_frame_uids]

        intervals = []
        interval_start_i = 0
        for i, frame_uid in enumerate(sorted_frame_uids[1:]):
            previous_frame_uid = sorted_frame_uids[i]

            if frame_uid - previous_frame_uid > 1:
                intervals.append(sorted_frame_uids[interval_start_i : i + 1])
                interval_start_i = i + 1

        intervals.append(sorted_frame_uids[interval_start_i : len(sorted_frame_uids)])
        interval_start_i = len(sorted_frame_uids)

        return intervals



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass

from ._object_annotation import _ObjectAnnotation
from .object import Object


@dataclass
class Seg3d(_ObjectAnnotation):
    """The 3D segmentation of a lidar pointcloud.

    Parameters
    ----------
    uid: str
        This a string representing the unique universal identifier of the annotation.
    point_ids: list of int
        The list of point indices.
    object: raillabel.format.Object
        A reference to the object, this annotation belongs to.
    sensor: raillabel.format.Sensor, optional
        A reference to the sensor, this annotation is labeled in. Default is None.
    attributes: dict, optional
        Attributes of the annotation. Dict keys are the name str of the attribute, values are the
        attribute values. Default is {}.

    Properties (read-only)
    ----------------------
    name: str
        Name of the annotation used by the VCD player for indexing in the object data pointers.
    """

    point_ids: t.List[int] = None

    OPENLABEL_ID = "vec"
    _REQ_FIELDS = ["point_ids"]

    @classmethod
    def fromdict(cls, data_dict: dict, sensors: dict, object: Object) -> "Seg3d":
        """Generate a Seg3d object from a dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        sensors: dict
            Dictionary containing all sensors for the scene.
        object: raillabel.format.Object
            Object this annotation belongs to.

        Returns
        -------
        annotation: Seg3d
            Converted annotation.
        """

        return Seg3d(
            uid=str(data_dict["uid"]),
            point_ids=data_dict["val"],
            object=object,
            sensor=cls._coordinate_system_fromdict(data_dict, sensors),
            attributes=cls._attributes_fromdict(data_dict),
        )

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = self._annotation_required_fields_asdict()

        dict_repr["val"] = [int(pid) for pid in self.point_ids]

        dict_repr.update(self._annotation_optional_fields_asdict())

        return dict_repr



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import IntrinsicsRadar

# == Fixtures =========================

@pytest.fixture
def intrinsics_radar_dict() -> dict:
    return {
        "resolution_px_per_m": 2.856,
        "width_px": 2856,
        "height_px": 1428,
    }

@pytest.fixture
def intrinsics_radar() -> dict:
    return IntrinsicsRadar(
        resolution_px_per_m=2.856,
        width_px=2856,
        height_px=1428,
    )

# == Tests ============================

def test_fromdict():
    intrinsics_radar = IntrinsicsRadar.fromdict(
        {
            "resolution_px_per_m": 2.856,
            "width_px": 2856,
            "height_px": 1428,
        }
    )

    assert intrinsics_radar.resolution_px_per_m == 2.856
    assert intrinsics_radar.width_px == 2856
    assert intrinsics_radar.height_px == 1428


def test_asdict():
    intrinsics_radar = IntrinsicsRadar(
        resolution_px_per_m=2.856,
        width_px=2856,
        height_px=1428,
    )

    assert intrinsics_radar.asdict() == {
        "resolution_px_per_m": 2.856,
        "width_px": 2856,
        "height_px": 1428,
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import Size3d

# == Fixtures =========================

@pytest.fixture
def size3d_dict() -> dict:
    return [0.35, 0.7, 1.92]

@pytest.fixture
def size3d() -> dict:
    return Size3d(0.35, 0.7, 1.92)

# == Tests ============================

def test_fromdict():
    size3d = Size3d.fromdict(
        [0.35, 0.7, 1.92]
    )

    assert size3d.x == 0.35
    assert size3d.y == 0.7
    assert size3d.z == 1.92


def test_asdict():
    size3d = Size3d(
        x=0.35,
        y=0.7,
        z=1.92
    )

    assert size3d.asdict() == [0.35, 0.7, 1.92]


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel._util._attribute_type import AttributeType
from raillabel.format import ElementDataPointer

# == Fixtures =========================

@pytest.fixture
def element_data_pointer_minimal_dict() -> dict:
    return {
        "type": "bbox",
        "frame_intervals": [],
        "attribute_pointers": {},
    }

@pytest.fixture
def element_data_pointer_minimal():
    return ElementDataPointer(
        uid="rgb_middle__bbox__person",
        frame_intervals=[],
        attribute_pointers={}
    )


@pytest.fixture
def element_data_pointer_full_dict(frame_interval_dict) -> dict:
    return {
        "type": "bbox",
        "frame_intervals": [
            frame_interval_dict
        ],
        "attribute_pointers": {
            "text_attr": "text",
            "num_attr": "num",
            "bool_attr": "boolean",
            "vec_attr": "vec",
        },
    }

@pytest.fixture
def element_data_pointer_full(sensor_camera, object_person, frame_interval):
    return ElementDataPointer(
        uid="rgb_middle__bbox__person",
        frame_intervals=[
            frame_interval
        ],
        attribute_pointers={
            "text_attr": AttributeType.TEXT,
            "num_attr": AttributeType.NUM,
            "bool_attr": AttributeType.BOOLEAN,
            "vec_attr": AttributeType.VEC,
        }
    )

# == Tests ============================


def test_asdict_minimal(sensor_camera, object_person):
    element_data_pointer = ElementDataPointer(
        uid="rgb_middle__bbox__person",
        frame_intervals=[],
        attribute_pointers={}
    )

    assert element_data_pointer.asdict() == {
        "type": "bbox",
        "frame_intervals": [],
        "attribute_pointers": {},
    }

def test_asdict_full(sensor_camera, object_person, frame_interval, frame_interval_dict):
    element_data_pointer = ElementDataPointer(
        uid="rgb_middle__bbox__person",
        frame_intervals=[
            frame_interval
        ],
        attribute_pointers={
            "text_attr": AttributeType.TEXT,
            "num_attr": AttributeType.NUM,
            "bool_attr": AttributeType.BOOLEAN,
            "vec_attr": AttributeType.VEC,
        }
    )

    assert element_data_pointer.asdict() == {
        "type": "bbox",
        "frame_intervals": [
            frame_interval_dict
        ],
        "attribute_pointers": {
            "text_attr": "text",
            "num_attr": "num",
            "bool_attr": "boolean",
            "vec_attr": "vec",
        },
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

import raillabel


@pytest.fixture
def loader():
    return raillabel.load_.loader_classes.LoaderRailLabel()


def test_supports_true(json_data, loader):
    assert loader.supports(json_data["openlabel_v1_short"])


def test_supports_false(json_data, loader):
    data = json_data["openlabel_v1_short"]
    data["openlabel"]["metadata"]["subschema_version"] = "4.0.0"
    assert not loader.supports(data)


# Tests the warnings and errors
def test_no_warnings(json_data, loader):
    loader.load(json_data["openlabel_v1_short"], validate=False)
    assert len(loader.warnings) == 0


def test_warnings_sync(json_data, loader):
    data = json_data["openlabel_v1_short"]

    data["openlabel"]["frames"]["0"]["frame_properties"]["streams"]["non_existing_stream"] = {
        "stream_properties": {
            "sync": {
                "timestamp": "1632321743.100000072"
            }
        }
    }

    loader.load(data, validate=False)
    assert len(loader.warnings) == 1

    # Tests for keywords in the warning that can help the user identify the source
    assert "frame" in loader.warnings[0]
    assert "0" in loader.warnings[0]
    assert "sync" in loader.warnings[0]
    assert "non_existing_stream" in loader.warnings[0]


def test_warnings_stream_sync_field(json_data, loader):
    data = json_data["openlabel_v1_short"]

    data["openlabel"]["frames"]["0"]["frame_properties"]["streams"][
        "rgb_middle"
    ]["stream_properties"]["stream_sync"] = data["openlabel"]["frames"]["0"][
        "frame_properties"
    ][
        "streams"
    ][
        "rgb_middle"
    ][
        "stream_properties"
    ][
        "sync"
    ]
    del data["openlabel"]["frames"]["0"]["frame_properties"]["streams"][
        "rgb_middle"
    ]["stream_properties"]["sync"]

    loader.load(data, validate=False)
    assert len(loader.warnings) == 1

    # Tests for keywords in the warning that can help the user identify the source
    assert "stream_sync" in loader.warnings[0]
    assert "deprecated" in loader.warnings[0].lower()
    assert "save()" in loader.warnings[0]


def test_identify_of_references(json_data, loader):
    data = json_data["openlabel_v1_short"]

    scene = loader.load(data, validate=False)

    for frame in scene.frames.values():

        for sensor_reference in frame.sensors.values():
            assert sensor_reference.sensor is scene.sensors[sensor_reference.sensor.uid]

        for frame_data in frame.frame_data.values():
            assert frame_data.sensor is scene.sensors[frame_data.sensor.uid]

        for annotation in frame.annotations.values():
            assert annotation.sensor is scene.sensors[annotation.sensor.uid]


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from decimal import Decimal

from ..format import Scene


def generate_timespan(scene: Scene) -> t.Tuple[t.Optional[Decimal], t.Optional[Decimal]]:
    """Return start and end timestamp of the scene.

    Parameters
    ----------
    scene: raillabel.format.Scene
        Scene the timespan should be based off.

    Returns
    -------
    decimal.Decimal or None
        Start timestamp of the scene. Is None if the scene has no frames.
    decimal.Decimal or None
        End timestamp of the scene. Is None if the scene has no frames.
    """
    start_timestamp = None
    end_timestamp = None

    for frame in scene.frames.values():

        if start_timestamp == None or frame.timestamp < start_timestamp:
            start_timestamp = frame.timestamp

        if end_timestamp == None or frame.timestamp > end_timestamp:
            end_timestamp = frame.timestamp

    return (start_timestamp, end_timestamp)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0


def _clean_dict(input_dict: dict) -> dict:
    """Remove all fields from a dict that are None or have a length of 0."""

    empty_keys = []
    for key, value in input_dict.items():

        is_field_empty = value is None or (hasattr(value, "__len__") and len(value) == 0)

        if is_field_empty:
            empty_keys.append(key)

    for key in empty_keys:
        del input_dict[key]

    return input_dict



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import Transform

# == Fixtures =========================

@pytest.fixture
def transform_dict(point3d_dict, quaternion_dict) -> dict:
    return {
        "translation": point3d_dict,
        "quaternion": quaternion_dict
    }

@pytest.fixture
def transform(point3d, quaternion) -> dict:
    return Transform(
        pos=point3d,
        quat=quaternion
    )

# == Tests ============================

def test_fromdict(point3d, point3d_dict, quaternion, quaternion_dict):
    transform = Transform.fromdict(
        {
            "translation": point3d_dict,
            "quaternion": quaternion_dict
        }
    )

    assert transform.pos == point3d
    assert transform.quat == quaternion


def test_asdict(point3d, point3d_dict, quaternion, quaternion_dict):
    transform = Transform(
        pos=point3d,
        quat=quaternion
    )

    assert transform.asdict() == {
        "translation": point3d_dict,
        "quaternion": quaternion_dict
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel


def test_load_raillabel(json_paths):
    data_path = json_paths["openlabel_v1_short"]
    scene = raillabel.load(data_path)
    assert len(scene.frames) != 0


def test_load_uai(json_paths):
    data_path = json_paths["understand_ai_t4_short"]
    scene = raillabel.load(data_path)
    assert len(scene.frames) != 0


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import logging
import typing as t
from io import StringIO


class _WarningsLogger:

    warnings: t.List[str] = []

    def __enter__(self) -> "_WarningsLogger":
        logger = logging.getLogger("loader_warnings")
        warnings_stream = StringIO()
        handler = logging.StreamHandler(warnings_stream)
        handler.setLevel(logging.WARNING)
        logger.addHandler(handler)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        logger = logging.getLogger("loader_warnings")
        stream = logger.handlers[-1].stream
        stream.seek(0)

        warnings_list = stream.getvalue().split("\n")

        if len(warnings_list) > 0:
            warnings_list = warnings_list[:-1]

        self.warnings = warnings_list


def _warning(message: str) -> logging.Logger:
    """Create a loader warning."""
    logging.getLogger("loader_warnings").warning(message)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import IntrinsicsPinhole

# == Fixtures =========================

@pytest.fixture
def intrinsics_pinhole_dict() -> dict:
    return {
        "camera_matrix": [
            0.48, 0   , 0.81, 0,
            0   , 0.16, 0.83, 0,
            0   , 0   , 1   , 0
        ],
        "distortion_coeffs": [
            0.49,
            0.69,
            0.31,
            0.81,
            0.99
        ],
        "width_px": 2464,
        "height_px": 1600
    }

@pytest.fixture
def intrinsics_pinhole() -> dict:
    return IntrinsicsPinhole(
        camera_matrix=(
            0.48, 0   , 0.81, 0,
            0   , 0.16, 0.83, 0,
            0   , 0   , 1   , 0
        ),
        distortion=(
            0.49,
            0.69,
            0.31,
            0.81,
            0.99
        ),
        width_px=2464,
        height_px=1600
    )

# == Tests ============================

def test_fromdict():
    intrinsics_pinhole = IntrinsicsPinhole.fromdict(
        {
            "camera_matrix": [
                0.48, 0   , 0.81, 0,
                0   , 0.16, 0.83, 0,
                0   , 0   , 1   , 0
            ],
            "distortion_coeffs": [
                0.49,
                0.69,
                0.31,
                0.81,
                0.99
            ],
            "width_px": 2464,
            "height_px": 1600
        }
    )

    assert intrinsics_pinhole.camera_matrix == (
        0.48, 0   , 0.81, 0,
        0   , 0.16, 0.83, 0,
        0   , 0   , 1   , 0
    )
    assert intrinsics_pinhole.distortion == (
        0.49,
        0.69,
        0.31,
        0.81,
        0.99
    )
    assert intrinsics_pinhole.width_px == 2464
    assert intrinsics_pinhole.height_px == 1600


def test_asdict():
    intrinsics_pinhole = IntrinsicsPinhole(
        camera_matrix=(
            0.48, 0   , 0.81, 0,
            0   , 0.16, 0.83, 0,
            0   , 0   , 1   , 0
        ),
        distortion=(
            0.49,
            0.69,
            0.31,
            0.81,
            0.99
        ),
        width_px=2464,
        height_px=1600
    )

    assert intrinsics_pinhole.asdict() == {
        "camera_matrix": [
            0.48, 0   , 0.81, 0,
            0   , 0.16, 0.83, 0,
            0   , 0   , 1   , 0
        ],
        "distortion_coeffs": [
            0.49,
            0.69,
            0.31,
            0.81,
            0.99
        ],
        "width_px": 2464,
        "height_px": 1600
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import Cuboid

# == Fixtures =========================

@pytest.fixture
def cuboid_dict(
    sensor_lidar,
    attributes_multiple_types_dict,
    point3d_dict,
    size3d_dict,
    quaternion_dict
) -> dict:
    return {
        "uid": "2c6b3de0-86c2-4684-b576-4cfd4f50d6ad",
        "name": "lidar__cuboid__person",
        "val": point3d_dict + quaternion_dict + size3d_dict,
        "coordinate_system": sensor_lidar.uid,
        "attributes": attributes_multiple_types_dict
    }

@pytest.fixture
def cuboid(
    point3d, size3d, quaternion,
    sensor_lidar,
    attributes_multiple_types,
    object_person
) -> dict:
    return Cuboid(
        uid="2c6b3de0-86c2-4684-b576-4cfd4f50d6ad",
        pos=point3d,
        quat=quaternion,
        size=size3d,
        object=object_person,
        sensor=sensor_lidar,
        attributes=attributes_multiple_types,
    )

# == Tests ============================

def test_fromdict(
    point3d, point3d_dict,
    size3d, size3d_dict,
    quaternion, quaternion_dict,
    sensor_lidar, sensors,
    object_person,
    attributes_multiple_types, attributes_multiple_types_dict,
):
    cuboid = Cuboid.fromdict(
        {
            "uid": "2c6b3de0-86c2-4684-b576-4cfd4f50d6ad",
            "name": "lidar__cuboid__person",
            "val": point3d_dict + quaternion_dict + size3d_dict,
            "coordinate_system": sensor_lidar.uid,
            "attributes": attributes_multiple_types_dict
        },
        sensors,
        object_person
    )

    assert cuboid.uid == "2c6b3de0-86c2-4684-b576-4cfd4f50d6ad"
    assert cuboid.name == "lidar__cuboid__person"
    assert cuboid.pos == point3d
    assert cuboid.quat == quaternion
    assert cuboid.size == size3d
    assert cuboid.object == object_person
    assert cuboid.sensor == sensor_lidar
    assert cuboid.attributes == attributes_multiple_types


def test_asdict(
    point3d, point3d_dict,
    size3d, size3d_dict,
    quaternion, quaternion_dict,
    sensor_lidar,
    object_person,
    attributes_multiple_types, attributes_multiple_types_dict,
):
    cuboid = Cuboid(
        uid="2c6b3de0-86c2-4684-b576-4cfd4f50d6ad",
        pos=point3d,
        quat=quaternion,
        size=size3d,
        object=object_person,
        sensor=sensor_lidar,
        attributes=attributes_multiple_types,
    )

    assert cuboid.asdict() == {
        "uid": "2c6b3de0-86c2-4684-b576-4cfd4f50d6ad",
        "name": "lidar__cuboid__person",
        "val": point3d_dict + quaternion_dict + size3d_dict,
        "coordinate_system": sensor_lidar.uid,
        "attributes": attributes_multiple_types_dict
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import Size2d

# == Fixtures =========================

@pytest.fixture
def size2d_dict() -> dict:
    return [25, 1.344]

@pytest.fixture
def size2d() -> dict:
    return Size2d(25, 1.344)

# == Tests ============================

def test_fromdict():
    size2d = Size2d.fromdict(
        [25, 1.344]
    )

    assert size2d.x == 25
    assert size2d.y == 1.344


def test_asdict():
    size2d = Size2d(
        x=25,
        y=1.344,
    )

    assert size2d.asdict() == [25, 1.344]


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])
